from keras_cv_attention_models.backend import layers, functional, models, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import add_pre_post_process
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.coco import eval_func, anchors_func
from keras_cv_attention_models.yolov8.yolov8 import YOLOV8Backbone, conv_bn, csp_with_2_conv, yolov8_head, switch_to_deploy

PRETRAINED_DICT = {
    "yolo_nas_l": {"coco": "cea25dd86e8e4aa1fd82b7e1288fa583"},
    "yolo_nas_m": {"coco": "16cbc4683b51894334b0264def1593a2"},
    "yolo_nas_s": {"coco": "283395afacb7ca5ea597d2e48dd19329"},
    "yolo_nas_l_before_reparam": {"coco": "a22ce5b0a036cf27a9e32fc80f09c373"},
    "yolo_nas_m_before_reparam": {"coco": "d093634fc5af38da0e763512ac829da4"},
    "yolo_nas_s_before_reparam": {"coco": "8c41ba972fc1bc66810c8f5ee15e404e"},
}


""" path aggregation fpn """


def upsample_merge(inputs, csp_depth=2, expansion=0.5, parallel_mode=False, use_alpha=True, use_reparam_conv=False, activation="relu", name=""):
    # print(f">>>> upsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    low, middle, high = inputs
    channels = high.shape[channel_axis]

    fpn_out = conv_bn(low, channels, activation=activation, name=name + "fpn_")
    low = layers.Conv2DTranspose(channels, kernel_size=2, strides=2, padding="valid", name=name + "up_conv_transpose")(fpn_out)

    middle = conv_bn(middle, channels, activation=activation, name=name + "middle_")
    high = conv_bn(high, channels, activation=activation, name=name + "high_")
    high = conv_bn(high, channels, kernel_size=3, strides=2, activation=activation, name=name + "high_down_")

    nn = functional.concat([low, middle, high], axis=channel_axis)
    nn = conv_bn(nn, channels, kernel_size=1, strides=1, activation=activation, name=name + "reduce_")
    shortcut, use_bias = True, True
    nn = csp_with_2_conv(nn, channels, csp_depth, shortcut, expansion, parallel_mode, use_bias, use_alpha, use_reparam_conv, activation=activation, name=name)
    return fpn_out, nn


def downsample_merge(inputs, csp_depth=2, expansion=0.5, parallel_mode=False, use_alpha=True, use_reparam_conv=False, activation="relu", name=""):
    # print(f">>>> downsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    # target_channel = inputs[-1].shape[channel_axis]
    inputs[0] = conv_bn(inputs[0], inputs[-1].shape[channel_axis], kernel_size=3, strides=2, activation=activation, name=name + "down_")
    nn = functional.concat(inputs, axis=channel_axis)
    channels = nn.shape[channel_axis]
    shortcut, use_bias = True, False
    nn = csp_with_2_conv(nn, channels, csp_depth, shortcut, expansion, parallel_mode, use_bias, use_alpha, use_reparam_conv, activation=activation, name=name)
    return nn


def path_aggregation_fpn(
    features, output_channels=[64, 128, 256], depthes=2, expansions=0.5, parallel_mode=False, use_alpha=True, use_reparam_conv=False, activation="relu", name=""
):
    # p5 ---+-> fpn_out0 -----------------------+-> pan_out0
    #       v [up -> concat]                    ^ [down -> concat]
    # p4 ---+-> f_out0 --------+-> fpn_out1 ----+-> pan_out1
    #       ^ [down -> concat] v [up -> concat] ^ [down -> concat]
    # p3 ---+------------------+----------------+-> pan_out2
    #                          ^ [down -> concat]
    # p2 ----------------------+
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    expansions = expansions if isinstance(expansions, (list, tuple)) else [expansions] * 4
    depthes = depthes if isinstance(depthes, (list, tuple)) else [depthes] * 4

    p2, p3, p4, p5 = features  # p2: [160, 160, 96], p3: [80, 80, 192], p4: [40, 40, 384], p5: [20, 20, 768]
    fpn_out0, f_out0 = upsample_merge([p5, p4, p3], depthes[0], expansions[0], parallel_mode, use_alpha, use_reparam_conv, activation, name=name + "p3p4p5_")
    fpn_out1, pan_out2 = upsample_merge(
        [f_out0, p3, p2], depthes[1], expansions[1], parallel_mode, use_alpha, use_reparam_conv, activation, name=name + "c3p3c2_"
    )

    use_reparam_conv = False  # False for downsample_merge blocks
    pan_out1 = downsample_merge([pan_out2, fpn_out1], depthes[2], expansions[2], parallel_mode, use_alpha, use_reparam_conv, activation, name=name + "c3n3_")
    pan_out0 = downsample_merge([pan_out1, fpn_out0], depthes[3], expansions[3], parallel_mode, use_alpha, use_reparam_conv, activation, name=name + "c3n4_")

    out = []
    for id, (pan_out, channel) in enumerate(zip([pan_out2, pan_out1, pan_out0], output_channels)):
        nn = conv_bn(pan_out, channel, activation=activation, name=name + "out{}_".format(id + 1))
        out.append(nn)
    return out


""" YOLO_NAS models """


def YOLO_NAS(
    backbone=None,
    csp_channels=[96, 192, 384, 768],  # [YOLOV8Backbone parameters]
    csp_depthes=[2, 3, 5, 2],
    features_pick=[-3, -2, -1],  # [Detector parameters]
    csp_expansions=[1 / 3, 1 / 3, 0.25, 0.25],
    paf_expansions=[1 / 3, 0.5, 1 / 3, 1 / 6],
    regression_len=68,  # bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64
    use_alpha=True,
    # use_bias=True,  # Actually used for reloading reparametered conv weights
    header_depth=1,
    use_reparam_conv=True,  # Use rep_vgg_conv_bn instead of conv_bn block in all csp_blocks.
    csp_parallel_mode=False,
    paf_parallel_mode=False,
    paf_output_channels=[64, 128, 256],
    paf_depthes=2,
    anchors_mode="yolov8",
    num_anchors="auto",  # "auto" means: anchors_mode=="anchor_free" / "yolov8" -> 1, anchors_mode=="yolor" -> 3, else 9
    use_object_scores="auto",  # "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False
    input_shape=(640, 640, 3),
    num_classes=80,
    activation="relu",
    classifier_activation="sigmoid",
    freeze_backbone=False,
    pretrained=None,
    model_name="yolo_nas",
    pyramid_levels_min=3,  # Init anchors for model prediction.
    anchor_scale="auto",  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
    rescale_mode="raw01",  # For decode predictions, raw01 means input value in range [0, 1].
    kwargs=None,  # Not using, recieving parameter
):
    if backbone is None:
        features_pick = [-4, -3, -2, -1]
        backbone_kwargs = dict(use_alpha=use_alpha, use_bias=True, use_reparam_conv=use_reparam_conv, input_shape=input_shape, activation=activation)
        backbone = YOLOV8Backbone(csp_channels, csp_depthes, features_pick, csp_expansions, csp_parallel_mode, **backbone_kwargs, model_name="backbone")
        features = backbone.outputs
    else:
        if isinstance(features_pick[0], str):
            features = [backbone.get_layer(layer_name) for layer_name in features_pick]
        else:
            features = model_surgery.get_pyramide_feature_layers(backbone)
            features = [features[id] for id in features_pick]
        feature_names, features = model_surgery.align_pyramide_feature_output_by_image_data_format(features)
        print(">>>> features:", {ii: jj.shape for ii, jj in zip(feature_names, features)})

    backbone.trainable = False if freeze_backbone else True
    use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, num_anchors, anchor_scale)
    inputs = backbone.inputs[0]

    fpn_features = path_aggregation_fpn(
        features, paf_output_channels, paf_depthes, paf_expansions, paf_parallel_mode, use_alpha, use_reparam_conv, activation=activation, name="pafpn_"
    )

    header_kwargs = {"depth": header_depth, "use_object_scores": use_object_scores, "activation": activation, "classifier_activation": classifier_activation}
    outputs = yolov8_head(fpn_features, num_classes, regression_len, num_anchors, hidden_channels=paf_output_channels, **header_kwargs, name="head_")
    outputs = layers.Activation("linear", dtype="float32", name="outputs_fp32")(outputs)

    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "yolov8", pretrained)

    pyramid_levels = [pyramid_levels_min, pyramid_levels_min + len(fpn_features) - 1]  # -> [3, 5], Note: path_aggregation_fpn merged one feature
    post_process = eval_func.DecodePredictions(
        backbone.input_shape[1:], pyramid_levels, anchors_mode, use_object_scores, anchor_scale, regression_len=regression_len
    )
    add_pre_post_process(model, rescale_mode=rescale_mode, post_process=post_process)
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


@register_model
def YOLO_NAS_S(
    input_shape=(640, 640, 3),
    freeze_backbone=False,
    num_classes=80,
    backbone=None,
    use_reparam_conv=True,
    classifier_activation="sigmoid",
    pretrained="coco",
    **kwargs
):
    return YOLO_NAS(**locals(), model_name=kwargs.pop("model_name", "yolo_nas_s_before_reparam" if use_reparam_conv else "yolo_nas_s"), **kwargs)


@register_model
def YOLO_NAS_M(
    input_shape=(640, 640, 3),
    freeze_backbone=False,
    num_classes=80,
    backbone=None,
    use_reparam_conv=True,
    classifier_activation="sigmoid",
    pretrained="coco",
    **kwargs
):
    csp_expansions = [2 / 3, 2 / 3, 2 / 3, 0.5]
    csp_parallel_mode = [True, True, True, False]
    paf_expansions = [1, 2 / 3, 1, 2 / 3]
    paf_output_channels = [96, 192, 384]
    paf_depthes = [2, 3, 2, 3]
    return YOLO_NAS(**locals(), model_name=kwargs.pop("model_name", "yolo_nas_m_before_reparam" if use_reparam_conv else "yolo_nas_m"), **kwargs)


@register_model
def YOLO_NAS_L(
    input_shape=(640, 640, 3),
    freeze_backbone=False,
    num_classes=80,
    backbone=None,
    use_reparam_conv=True,
    classifier_activation="sigmoid",
    pretrained="coco",
    **kwargs
):
    csp_expansions = [1, 2 / 3, 2 / 3, 2 / 3]
    csp_parallel_mode = True
    paf_expansions = [2 / 3, 4 / 3, 2 / 3, 2 / 3]
    paf_output_channels = [128, 256, 512]
    paf_depthes = 4
    return YOLO_NAS(**locals(), model_name=kwargs.pop("model_name", "yolo_nas_l_before_reparam" if use_reparam_conv else "yolo_nas_l"), **kwargs)



from keras_cv_attention_models.yolor.yolor import (
    CSPDarknet,
    YOLOR,
    YOLOR_CSP,
    YOLOR_CSPX,
    YOLOR_P6,
    YOLOR_W6,
    YOLOR_E6,
    YOLOR_D6,
)

__head_doc__ = """
Keras implementation of [Github WongKinYiu/yolor](https://github.com/WongKinYiu/yolor).
Paper [Paper 2105.04206 You Only Learn One Representation: Unified Network for Multiple Tasks](https://arxiv.org/pdf/2105.04206.pdf).

Args:
  backbone: backbone model, could be any model with pyramid stage structure.
      Default None for CSPDarknet with csp_depthes={csp_depthes}, csp_channels={csp_channels}.
"""

__tail_doc__ = """  features_pick: specific `layer names` or `pyramid feature indexes` from backbone model.
        Default `[-3, -2, -1]` means using the last 3 pyramid feature output from backbone.
  regression_len: bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64.
  anchors_mode: one of ["efficientdet", "anchor_free", "yolor"], controls which anchor to use.
      - efficientdet anchors default settings: use_object_scores=False, num_anchors=9, anchor_scale=4,
          aspect_ratios=[1, 2, 0.5], num_scales=3, grid_zero_start=False.
      - anchor_free default settings: use_object_scores=True, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=True.
      - yolor default settings: use_object_scores=True, num_anchors=3.
      - yolov8 default settings: use_object_scores=False, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=False.
      Default "yolor".
  num_anchors: number of anchors for a single grid point, should be same with dataset used value.
      Default "auto" means: anchors_mode=="anchor_free" / "yolov8" -> 1, anchors_mode=="yolor" -> 3, else 9.
  use_object_scores: bollean value if model header output includes `object_scores`.
      Default "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False.
  input_shape: input shape if backbone is None, else will use input_shape from backbone.
  num_classes: total output classes. Set `0` to disable `classifier` output. Default 80 for COCO.
  activation: activation used in whole model, default `swish`. Default "swish".
  classifier_activation: The activation function to use for classifier output if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer. Default `sigmoid`.
  freeze_backbone: set `True` for `backbone.trainable = False`. Default `False`.
  pretrained: one of `None` (random initialization) or 'coco' (pre-training on COCO).
      Will try to download and load pre-trained model weights if not None. Default `coco`.
  pyramid_levels_min: anchors inititial parameter for model prediction, not affecting model architecture. Default `3`.
      pyramid_levels_max is calculated as `pyramid_levels_min + len(features_pick) - 1`.
  anchor_scale: anchors inititial parameter for model prediction, not affecting model architecture.
      Default "auto" means: 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4.
  rescale_mode: model precessing input, not for model structure. Defulat "raw01" means input value in range `[0, 1]`.

Returns:
    A `keras.Model` instance.
"""

YOLOR.__doc__ = __head_doc__.format(csp_depthes=[2, 8, 8, 4], csp_channels=[128, 256, 512, 1024]) + """  csp_depthes: CSPDarknet backbone depth for each block, default `[2, 8, 8, 4]`.
  csp_channels: CSPDarknet backbone channel for each block, default: `[128, 256, 512, 1024]`.
  stem_width: CSPDarknet backbone stem width, default -1 means using csp_channels[0] // 2.
  use_focus_stem: boolean value if CSPDarknet backbone using focus_stem or conv one, default False.
  ssp_depth: CSPDarknet backbone spatial_pyramid_pooling depth, default 2.
  csp_use_pre: boolean value if CSPDarknet backbone blocks using pre-conv for deep branch, default False.
  csp_use_post: boolean value if CSPDarknet backbone blocks using post-conv for deep branch, default True.
  use_csp_downsample: boolean value if CSPDarknet backbone and FPN downsample using `csp_conv_downsample`
      or a conv layer with `kernel_size=3, strides=2`. Default False.
  fpn_depth: depth for FPN headers, default 2.
  use_depthwise_conv: boolean value if using additional depthwise conv.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model      | Params | FLOPs   | Input | COCO val AP | test AP |
  | ---------- | ------ | ------- | ----- | ----------- | ------- |
  | YOLOR_CSP  | 52.9M  | 60.25G  | 640   | 50.0        | 52.8    |
  | YOLOR_CSPX | 99.8M  | 111.11G | 640   | 51.5        | 54.8    |
  | YOLOR_P6   | 37.3M  | 162.87G | 1280  | 52.5        | 55.7    |
  | YOLOR_W6   | 79.9M  | 226.67G | 1280  | 53.6 ?      | 56.9    |
  | YOLOR_E6   | 115.9M | 341.62G | 1280  | 50.3 ?      | 57.6    |
  | YOLOR_D6   | 151.8M | 467.88G | 1280  | 50.8 ?      | 58.2    |
"""

YOLOR_CSP.__doc__ = __head_doc__.format(csp_depthes=[2, 8, 8, 4], csp_channels=[128, 256, 512, 1024]) + __tail_doc__
YOLOR_CSP.__doc__ = __head_doc__.format(csp_depthes=[3, 10, 10, 5], csp_channels=[160, 320, 640, 1280]) + __tail_doc__
YOLOR_P6.__doc__ = __head_doc__.format(csp_depthes=[3, 7, 7, 3, 3], csp_channels=[128, 256, 384, 512, 640]) + __tail_doc__
YOLOR_W6.__doc__ = __head_doc__.format(csp_depthes=[3, 7, 7, 3, 3], csp_channels=[128, 256, 512, 768, 1024]) + __tail_doc__
YOLOR_E6.__doc__ = __head_doc__.format(csp_depthes=[3, 7, 7, 3, 3], csp_channels=[160, 320, 640, 960, 1280]) + __tail_doc__
YOLOR_D6.__doc__ = __head_doc__.format(csp_depthes=[3, 15, 15, 7, 7], csp_channels=[160, 320, 640, 960, 1280]) + __tail_doc__



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    conv2d_no_bias,
    layer_norm,
    mlp_block,
    mlp_block_with_depthwise_conv,
    multi_head_self_attention,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "caformer_s18": {
        "imagenet": {224: "b8a824d4197161286de2aef08c1be379", 384: "fcd11441f1811124f77a4cad1b328639"},
        "imagenet21k-ft1k": {224: "4b87b7d0393dc607089eff549dfe7319", 384: "05f20bcd5403a6076b1fa3276766b987"},
    },
    "caformer_s36": {
        "imagenet": {224: "df683079147328b5cafeb6e98e99a885", 384: "532becc7aaa087430baa1691ed2e64eb"},
        "imagenet21k-ft1k": {224: "89b40eef0fc073932460e2ba8094acb7", 384: "f4d163c51c8f06f452cd0f312ee51c3d"},
    },
    "caformer_m36": {
        "imagenet": {224: "04a88000f78cc8c1e9b261dc71fa3e6f", 384: "3c3b13ffb3cfbd304897abaef5fb0abf"},
        "imagenet21k-ft1k": {224: "af5a2978cc60609ed70941f506d4a337", 384: "3055798b1c0f291547911460390c1964"},
    },
    "caformer_b36": {
        "imagenet": {224: "8461e68f9dde4fe6ef201874e4d6cb66", 384: "087bddb30c6f04a44e4b3a08864f2327"},
        "imagenet21k-ft1k": {224: "47f77c5087b4acd427ec960074136644", 384: "a6c78cb35254870a475487b4a582bcd1"},
    },
    "convformer_s18": {
        "imagenet": {224: "ef64c846c5890258227281fc6d2dedf7", 384: "aa8d9c6e6f47e3934ed329ded7f85674"},
        "imagenet21k-ft1k": {224: "da24d177fc47d878ae3621be82130ede", 384: "1e8ba700dc67840ecde6c76f0688d2f4"},
    },
    "convformer_s36": {
        "imagenet": {224: "2eee80c9126a08cb09c386de7d9759b6", 384: "a8f175274494598cb322ed0c5f17e4b9"},
        "imagenet21k-ft1k": {224: "e75e739dc0f09dbd88b326520e4b7fb0", 384: "2f9bfe0664ff3128820d2873915709a1"},
    },
    "convformer_m36": {
        "imagenet": {224: "2e7a8c5f9827b0ca8fe51f9a0aa92223", 384: "3a53ed969febefb1d080ce0b78dcbb33"},
        "imagenet21k-ft1k": {224: "652c8ec792c50a0f87471c5d53e92940", 384: "f37c31964b650a7700e7820ea4ccf69e"},
    },
    "convformer_b36": {
        "imagenet": {224: "c96d0f4720c36ae19ab8eee02c6e6034", 384: "109236ff75aabe1885ef625c3bfe756c"},
        "imagenet21k-ft1k": {224: "64774d61660d2e95df8d274785f7708a", 384: "b7e350d127a37ddfa7cd611ce21b4e4b"},
    },
}


def meta_former_block(inputs, use_attn=False, head_dim=32, mlp_ratio=4, layer_scale=0, residual_scale=0, drop_rate=0, activation="star_relu", name=""):
    # channnel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[-1]

    """ attention """
    nn = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, center=False, axis=-1, name=name + "attn_")
    # nn = conv_pool_attention_mixer(nn, num_heads, num_attn_low_heads=num_attn_low_heads, pool_size=pool_size, activation=activation, name=name + "attn_")
    if use_attn:
        nn = multi_head_self_attention(nn, num_heads=input_channel // head_dim, name=name + "mhsa_")
    else:
        nn = mlp_block_with_depthwise_conv(nn, input_channel * 2, kernel_size=7, use_bias=False, activation=(activation, None), name=name + "mlp_sep_")
    attn_out = add_with_layer_scale_and_drop_block(
        inputs, nn, layer_scale=layer_scale, residual_scale=residual_scale, drop_rate=drop_rate, axis=-1, name=name + "attn_"
    )

    """ MLP """
    nn = layer_norm(attn_out, epsilon=LAYER_NORM_EPSILON, center=False, axis=-1, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, use_bias=False, activation=activation, name=name + "mlp_")
    nn = add_with_layer_scale_and_drop_block(
        attn_out, nn, layer_scale=layer_scale, residual_scale=residual_scale, drop_rate=drop_rate, axis=-1, name=name + "mlp_"
    )
    return nn


def CAFormer(
    num_blocks=[3, 3, 9, 3],
    out_channels=[64, 128, 320, 512],
    block_types=["conv", "conv", "transform", "transform"],
    head_dim=32,
    mlp_ratios=4,
    head_filter=2048,
    head_filter_activation="squared_relu",
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="star_relu",
    drop_connect_rate=0,
    dropout=0,
    layer_scales=0,
    residual_scales=[0, 0, 1, 1],
    classifier_activation="softmax",
    pretrained=None,
    model_name="caformer",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = layers.ZeroPadding2D(padding=2, name="stem_")(inputs)  # padding=2
    nn = conv2d_no_bias(nn, out_channels[0], kernel_size=7, strides=4, padding="valid", use_bias=True, name="stem_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1], name="stem_pre_permute")(nn)
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, center=False, axis=-1, name="stem_")

    """ stacks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, block_type) in enumerate(zip(num_blocks, out_channels, block_types)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, center=False, axis=-1, name=stack_name + "downsample_")
            nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2], name=stack_name + "permute_pre")(nn)
            nn = conv2d_no_bias(nn, out_channel, 3, strides=2, padding="same", use_bias=True, name=stack_name + "downsample_")
            nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1], name=stack_name + "permute_post")(nn)

        use_attn = True if block_type[0].lower() == "t" else False
        if use_attn:
            block_height, block_width = nn.shape[1:-1]
            nn = functional.reshape(nn, [-1, block_height * block_width, nn.shape[-1]])  # Using 3D for attention inputs

        mlp_ratio = mlp_ratios[stack_id] if isinstance(mlp_ratios, (list, tuple)) else mlp_ratios
        layer_scale = layer_scales[stack_id] if isinstance(layer_scales, (list, tuple)) else layer_scales
        residual_scale = residual_scales[stack_id] if isinstance(residual_scales, (list, tuple)) else residual_scales
        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = meta_former_block(nn, use_attn, head_dim, mlp_ratio, layer_scale, residual_scale, block_drop_rate, activation=activation, name=name)
            global_block_id += 1

        if use_attn:
            nn = functional.reshape(nn, [-1, block_height, block_width, nn.shape[-1]])  # Revert 3D to 4D
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2], name="pre_output_permute")(nn)

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="pre_output_")
        if head_filter > 0:
            nn = layers.Dense(head_filter, use_bias=True, name="feature_dense")(nn)
            head_filter_activation = head_filter_activation if head_filter_activation is not None else activation
            nn = activation_by_name(nn, activation=head_filter_activation, name="feature_")
            nn = layer_norm(nn, name="feature_")  # epsilon=1e-5
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "caformer", pretrained)
    return model


@register_model
def CAFormerS18(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    kwargs.pop("kwargs", None)
    return CAFormer(**locals(), model_name=kwargs.pop("model_name", "caformer_s18"), **kwargs)


@register_model
def CAFormerS36(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 12, 18, 3]
    kwargs.pop("kwargs", None)
    return CAFormer(**locals(), model_name=kwargs.pop("model_name", "caformer_s36"), **kwargs)


@register_model
def CAFormerM36(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 12, 18, 3]
    out_channels = [96, 192, 384, 576]
    head_filter = out_channels[-1] * 4
    kwargs.pop("kwargs", None)
    return CAFormer(**locals(), model_name=kwargs.pop("model_name", "caformer_m36"), **kwargs)


@register_model
def CAFormerB36(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 12, 18, 3]
    out_channels = [128, 256, 512, 768]
    head_filter = out_channels[-1] * 4
    kwargs.pop("kwargs", None)
    return CAFormer(**locals(), model_name=kwargs.pop("model_name", "caformer_b36"), **kwargs)


""" ConvFormer """


@register_model
def ConvFormerS18(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return CAFormerS18(**locals(), block_types=["conv", "conv", "conv", "conv"], model_name="convformer_s18", **kwargs)


@register_model
def ConvFormerS36(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return CAFormerS36(**locals(), block_types=["conv", "conv", "conv", "conv"], model_name="convformer_s36", **kwargs)


@register_model
def ConvFormerM36(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return CAFormerM36(**locals(), block_types=["conv", "conv", "conv", "conv"], model_name="convformer_m36", **kwargs)


@register_model
def ConvFormerB36(input_shape=(224, 224, 3), num_classes=1000, activation="star_relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return CAFormerB36(**locals(), block_types=["conv", "conv", "conv", "conv"], model_name="convformer_b36", **kwargs)



import keras_core
from keras_core.ops import *
from keras_core.ops import concatenate as concat
from keras_core.ops import mean as reduce_mean
from keras_core.ops import max as reduce_max
from keras_core.ops import min as reduce_min
from keras_core.ops import power as pow
from keras_core.ops import clip as clip_by_value


def extract_patches(images, sizes=1, strides=1, rates=1, padding="valid", name=None):
    return keras_core.ops.image.extract_patches(
        images,
        size=sizes[1:-1] if isinstance(sizes, int) or len(sizes) > 2 else sizes,
        strides=strides[1:-1] if isinstance(strides, int) or len(strides) > 2 else strides,
        dilation_rate=rates[1:-1] if isinstance(rates, int) or len(rates) > 2 else rates,
        padding=padding.lower(),
        data_format=keras_core.backend.image_data_format(),
    )


def gather(inputs, indices, axis=None, batch_dims=0, name=None):
    """Defaults axis=None means the first non-batch dimension"""
    axis = batch_dims if axis is None else (len(inputs.shape) + axis if axis < 0 else axis)
    return keras_core.ops.take(inputs, indices, axis=axis)


def l2_normalize(inputs, axis=None, epsilon=1e-12, name=None):
    return inputs / keras_core.ops.sqrt(keras_core.ops.maximum(keras_core.ops.sum(inputs**2, axis=axis, keepdims=True), epsilon))


def norm(inputs, ord="euclidean", axis=1, keepdims=False, name=None):
    return keras_core.ops.sqrt(keras_core.ops.sum(inputs**2, axis=axis, keepdims=True))


def resize(images, size, method="bilinear", preserve_aspect_ratio=False, antialias=False, name=None):
    return keras_core.ops.image.resize(images, size, interpolation=method, antialias=antialias, data_format=keras_core.backend.image_data_format())


def reduce_sum(inputs, axis=None, keepdims=False, name=None):
    axis = () if axis is None else axis
    if isinstance(inputs, (list, tuple)) and axis == 0:
        rr = inputs[0]
        for ii in inputs[1:]:
            rr += ii
        return rr
    else:
        # return wrapper(lambda xx: xx.sum(dim=axis, keepdim=keepdims), inputs, name=name)
        return keras_core.ops.sum(inputs, axis=axis, keepdims=keepdims)


def rsqrt(inputs, name=None):
    return keras_core.ops.true_divide(1, keras_core.ops.sqrt(inputs))


def split(inputs, num_or_size_splits, axis=0, num=None, name="split"):
    from builtins import sum

    if isinstance(num_or_size_splits, int):
        return keras_core.ops.split(inputs, num_or_size_splits, axis=axis)

    axis = (len(inputs.shape) + axis) if axis < 0 else axis
    split_axis_shape = inputs.shape[axis]
    assert split_axis_shape is not None

    size_splits = num_or_size_splits
    size_splits = [0 if ii is None or ii == -1 else ii for ii in size_splits]
    num_unknown_dim = sum([ii == 0 for ii in size_splits])
    assert num_unknown_dim < 2, "At most one unknown dimension in num_or_size_splits: {}".format(num_or_size_splits)

    if num_unknown_dim == 1:
        size_splits = [(split_axis_shape - sum(size_splits)) if ii == 0 else ii for ii in size_splits]

    cum_split = [sum(size_splits[: id + 1]) for id, _ in enumerate(size_splits)]
    # len(keras_core.ops.split(np.ones([2, 6]), [2, 2, 2], axis=-1)) == 4
    # len(keras_core.ops.split(keras_core.layers.Input([6]), [2, 2, 2], axis=-1)) == 3
    return keras_core.ops.split(inputs, cum_split, axis=axis)[: len(size_splits)]



import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    ChannelAffine,
    depthwise_conv2d_no_bias,
    drop_block,
    drop_connect_rates_split,
    group_norm,
    layer_norm,
    make_divisible,
    mlp_block,
    multi_head_self_attention,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-5
PRETRAINED_DICT = {
    "mobilevit_xxs": {"imagenet": "f9d1d4f7329726b4bb9069cd631a152e"},
    "mobilevit_xs": {"imagenet": "dcd8591668946aa4ddb0159dfe65cc2b"},
    "mobilevit_s": {"imagenet": "55f1051035ecd14e90ae3df80fd0c7f3"},
    "mobilevit_v2_050": {"imagenet": {256: "a842a40c0f49dc2bbe935493caed061b"}},
    "mobilevit_v2_075": {"imagenet": {256: "8588b3d6bf4aa750766ddc6d01824c67"}},
    "mobilevit_v2_100": {"imagenet": {256: "55d499bbc29f0f6379a4cc6f610e10e8"}},
    "mobilevit_v2_125": {"imagenet": {256: "b8af7b7668774796530f19dd5b6080fb"}},
    "mobilevit_v2_150": {
        "imagenet": {256: "065e7a07f7e2d0d74a33913195df9044"},
        "imagenet22k": {256: "cf3c4ec278154ece62e8967faa5c0391", 384: "cdcfaebb573f8cd1f41044ac0e958204"},
    },
    "mobilevit_v2_175": {
        "imagenet": {256: "627719428c6cb35f071a7ea69a6961c4"},
        "imagenet22k": {256: "b849708a6b2c1f115b8b8c366e1d1a19", 384: "d3feef5108b6195d1c5525fb185bf720"},
    },
    "mobilevit_v2_200": {
        "imagenet": {256: "1fe59d8bb2662761084d1c04259a778d"},
        "imagenet22k": {256: "931f0be1761bcf8443359ec1661bb6a7", 384: "1dc6cdafb187611e5a4819272d64fba7"},
    },
}


def bottle_in_linear_out_block(inputs, out_channel, strides=1, expand_ratio=4, use_shortcut=False, drop_rate=0, activation="swish", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    hidden_dim = int(input_channel * expand_ratio)
    deep = conv2d_no_bias(inputs, hidden_dim, kernel_size=1, strides=1, name=name + "deep_1_")
    deep = batchnorm_with_activation(deep, activation=activation, name=name + "deep_1_")
    deep = depthwise_conv2d_no_bias(deep, kernel_size=3, strides=strides, padding="same", name=name + "deep_2_")
    deep = batchnorm_with_activation(deep, activation=activation, name=name + "deep_2_")
    deep = conv2d_no_bias(deep, out_channel, kernel_size=1, strides=1, name=name + "deep_3_")
    deep = batchnorm_with_activation(deep, activation=None, name=name + "deep_3_")
    deep = drop_block(deep, drop_rate=drop_rate, name=name + "deep_")

    out = layers.Add()([inputs, deep]) if use_shortcut else deep
    return out


def linear_self_attention(inputs, qkv_bias=False, out_bias=False, attn_axis=2, attn_dropout=0, name=None):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    qkv = conv2d_no_bias(inputs, 1 + input_channel * 2, kernel_size=1, use_bias=qkv_bias, name=name and name + "qkv_")
    query, key, value = functional.split(qkv, [1, input_channel, input_channel], axis=channel_axis)
    context_score = layers.Softmax(axis=attn_axis, name=name and name + "attention_scores")(query)  # on patch_hh * patch_ww dimension
    context_score = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(context_score) if attn_dropout > 0 else context_score
    # print(f"{query.shape = }, {key.shape = }, {value.shape = }, {context_score.shape = }")

    context_vector = layers.Multiply()([key, context_score])  # [batch, height, width, input_channel]
    context_vector = functional.reduce_sum(context_vector, keepdims=True, axis=attn_axis)  # on patch_hh * patch_ww dimension

    out = functional.relu(value) * context_vector
    out = conv2d_no_bias(out, input_channel, kernel_size=1, use_bias=out_bias, name=name and name + "output")
    return out


def mhsa_mlp_block(
    inputs,
    out_channel,
    num_heads=4,
    qkv_bias=True,
    mlp_ratio=4,
    num_norm_groups=-1,  # -1 or 0 for V1 using layer_norm, or 1 for V2 using group_norm
    mlp_drop_rate=0,
    attn_drop_rate=0,
    drop_rate=0,
    layer_scale=-1,
    activation="gelu",
    name=None,
):
    # print(f"{inputs.shape = }")
    attn = group_norm(inputs, groups=num_norm_groups, axis=-1, name=name + "attn_") if num_norm_groups > 0 else layer_norm(inputs, axis=-1, name=name + "attn_")
    attn = multi_head_self_attention(attn, num_heads, qkv_bias=qkv_bias, out_bias=True, attn_dropout=attn_drop_rate, name=name and name + "attn_mhsa_")
    attn = add_with_layer_scale_and_drop_block(inputs, attn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name and name + "1_")

    mlp = group_norm(attn, groups=num_norm_groups, axis=-1, name=name + "mlp_") if num_norm_groups > 0 else layer_norm(attn, axis=-1, name=name + "mlp_")
    mlp = mlp_block(mlp, int(out_channel * mlp_ratio), drop_rate=mlp_drop_rate, activation=activation, name=name and name + "mlp_")
    return add_with_layer_scale_and_drop_block(attn, mlp, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name and name + "2_")


def linear_mhsa_mlp_block(
    inputs,
    out_channel,
    num_heads=4,
    qkv_bias=True,
    mlp_ratio=4,
    num_norm_groups=-1,  # -1 or 0 for V1 using layer_norm, or 1 for V2 using group_norm
    mlp_drop_rate=0,
    attn_drop_rate=0,
    drop_rate=0,
    layer_scale=-1,
    activation="gelu",
    name=None,
):
    attn = group_norm(inputs, groups=num_norm_groups, name=name + "attn_") if num_norm_groups > 0 else layer_norm(inputs, name=name + "attn_")
    attn = layers.Reshape(attn.shape[1:])(attn)  # Or will throw error when converting tflite, if GroupNorm is followed by Conv2D
    attn = linear_self_attention(attn, qkv_bias=qkv_bias, out_bias=True, attn_dropout=attn_drop_rate, name=name and name + "attn_mhsa_")
    attn = add_with_layer_scale_and_drop_block(inputs, attn, layer_scale=layer_scale, drop_rate=drop_rate, name=name and name + "1_")

    mlp = group_norm(attn, groups=num_norm_groups, name=name + "mlp_") if num_norm_groups > 0 else layer_norm(attn, name=name + "mlp_")
    mlp = layers.Reshape(mlp.shape[1:])(mlp)  # Or will throw error when converting tflite, if GroupNorm is followed by Conv2D
    mlp = mlp_block(mlp, int(out_channel * mlp_ratio), drop_rate=mlp_drop_rate, use_conv=True, activation=activation, name=name and name + "mlp_")
    return add_with_layer_scale_and_drop_block(attn, mlp, layer_scale=layer_scale, drop_rate=drop_rate, name=name and name + "2_")


def transformer_pre_process(inputs, out_channel, patch_size=2, resize_first=False, use_depthwise=False, patches_to_batch=True, activation="swish", name=""):
    height_axis, width_axis, channel_axis = (1, 2, 3) if image_data_format() == "channels_last" else (2, 3, 1)
    nn = inputs

    if resize_first:  # V2
        patch_hh, patch_ww = int(math.ceil(nn.shape[height_axis] / patch_size)), int(math.ceil(nn.shape[width_axis] / patch_size))
        # print(f"transformer_pre_process before resize: {nn.shape = }")
        if patch_hh * patch_size != nn.shape[height_axis] or patch_ww * patch_size != nn.shape[width_axis]:
            nn = functional.resize(nn, [patch_hh * patch_size, patch_ww * patch_size], method="bilinear")

    if use_depthwise:  # V2
        nn = depthwise_conv2d_no_bias(nn, kernel_size=3, strides=1, padding="same", name=name + "pre_1_")
    else:  # V1
        nn = conv2d_no_bias(nn, nn.shape[channel_axis], kernel_size=3, strides=1, padding="same", name=name + "pre_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "pre_1_")
    nn = conv2d_no_bias(nn, out_channel, kernel_size=1, strides=1, name=name + "pre_2_")

    if not resize_first:  # V1
        patch_hh, patch_ww = int(math.ceil(nn.shape[height_axis] / patch_size)), int(math.ceil(nn.shape[width_axis] / patch_size))
        # print(f"transformer_pre_process before resize: {nn.shape = }")
        if patch_hh * patch_size != nn.shape[height_axis] or patch_ww * patch_size != nn.shape[width_axis]:
            nn = functional.resize(nn, [patch_hh * patch_size, patch_ww * patch_size], method="bilinear")

    # Extract patchs, limit transpose permute length <= 4
    # [batch, height, width, channel] -> [batch, height // 2, 2, width // 2, 2, channel] -> [batch * 4, height // 2, width // 2, channel]
    # print(f"transformer_pre_process after resize: {nn.shape = }")
    if image_data_format() == "channels_last":
        nn = functional.reshape(nn, [-1, patch_ww, patch_size, out_channel])  # [B * patch_hh * h_patch_size, patch_ww, w_patch_size, C]
        nn = functional.transpose(nn, [0, 2, 1, 3])  # [B * patch_hh * h_patch_size, w_patch_size, patch_ww, C]
        nn = functional.reshape(nn, [-1, patch_hh, patch_size * patch_size, patch_ww * out_channel])  # [B, patch_hh, h_patch_size * w_patch_size, patch_ww * C]
        nn = functional.transpose(nn, [0, 2, 1, 3])  # [B, h_patch_size * w_patch_size, patch_hh, patch_ww * C]
        extract_shape = [-1, patch_hh, patch_ww, out_channel] if patches_to_batch else [-1, patch_size * patch_size, patch_hh * patch_ww, out_channel]
        nn = functional.reshape(nn, extract_shape)  # channels_last
    else:  # [B, C, patch_hh * h_patch_size, patch_ww * w_patch_size]
        nn = functional.reshape(nn, [-1, patch_size, patch_ww, patch_size])  # [B * C * patch_hh, h_patch_size, patch_ww, w_patch_size]
        nn = functional.transpose(nn, [0, 2, 1, 3])  # [B * C * patch_hh, patch_ww, h_patch_size, w_patch_size]
        nn = functional.reshape(nn, [-1, out_channel, patch_hh * patch_ww, patch_size * patch_size])  # [B, C, patch_hh * patch_ww, h_patch_size * w_patch_size]
        if patches_to_batch:  # V1 -> channels_last, V2 keep channels_first, `patch_hh * patch_ww` -> attn_axis=2
            nn = functional.transpose(nn, [0, 3, 2, 1])  # [B, h_patch_size * w_patch_size, patch_hh * patch_ww, C]
            nn = functional.reshape(nn, [-1, patch_hh, patch_ww, out_channel])  # channels_last
    return nn


def transformer_post_process(inputs, pre_attn, out_channel, patch_size=2, patch_height=-1, activation="swish", name=""):
    height_axis, width_axis, channel_axis = (1, 2, 3) if image_data_format() == "channels_last" else (2, 3, 1)
    patches_to_batch = patch_height == -1
    if patches_to_batch:  # V1, [batch * 4, height // 2, width // 2, channel], channels_last
        patch_hh, patch_ww, channel = inputs.shape[1], inputs.shape[2], inputs.shape[3]
    else:  # V2, [batch, 4, height // 2 * width // 2, channel], channels_last or channels_first
        patch_hh, patch_ww, channel = patch_height, inputs.shape[2] // patch_height, inputs.shape[channel_axis]
    # print(f"{patch_hh = }, {patch_ww = }, {channel = }, {inputs.shape = }")

    # [B * 4, height // 2, width // 2, C] -> [B, height // 2, 2, width // 2, 2, C] -> [B, height, width, C]
    if image_data_format() == "channels_last":
        nn = functional.reshape(inputs, [-1, patch_size * patch_size, patch_hh, patch_ww * channel])  # [B, h_patch_size * w_patch_size, patch_hh, patch_ww * C]
        nn = functional.transpose(nn, [0, 2, 1, 3])  # [B, patch_hh, h_patch_size * w_patch_size, patch_ww * C]
        nn = functional.reshape(nn, [-1, patch_size, patch_ww, channel])  # [B * patch_hh * h_patch_size, w_patch_size, patch_ww, C]
        nn = functional.transpose(nn, [0, 2, 1, 3])  # [B * patch_hh * h_patch_size, patch_ww, w_patch_size, C]
        nn = functional.reshape(nn, [-1, patch_hh * patch_size, patch_ww * patch_size, channel])
    else:
        nn = inputs
        if patches_to_batch:
            nn = functional.reshape(nn, [-1, patch_size * patch_size, patch_hh * patch_ww, channel])  # [B, h_patch_size * w_patch_size, patch_hh * patch_ww, C]
            nn = functional.transpose(nn, [0, 3, 2, 1])  # [B, C, patch_hh * patch_ww, h_patch_size * w_patch_size]
        nn = functional.reshape(nn, [-1, patch_ww, patch_size, patch_size])  # [B * C * patch_hh, patch_ww, h_patch_size, w_patch_size]
        nn = functional.transpose(nn, [0, 2, 1, 3])  # [B * C * patch_hh, h_patch_size, patch_ww, w_patch_size]
        nn = functional.reshape(nn, [-1, channel, patch_hh * patch_size, patch_ww * patch_size])

    # print(f"transformer_post_process before resize: {nn.shape = }")
    if pre_attn is not None and (nn.shape[height_axis] != pre_attn.shape[height_axis] or nn.shape[width_axis] != pre_attn.shape[width_axis]):
        nn = functional.resize(nn, [pre_attn.shape[height_axis], pre_attn.shape[width_axis]], method="bilinear")
    # print(f"transformer_post_process after resize: {nn.shape = }")

    nn = conv2d_no_bias(nn, out_channel, kernel_size=1, strides=1, name=name + "post_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "post_1_")
    if pre_attn is not None:  # V1
        nn = functional.concat([pre_attn, nn], axis=channel_axis)
        nn = conv2d_no_bias(nn, out_channel, kernel_size=3, strides=1, padding="same", name=name + "post_2_")
        nn = batchnorm_with_activation(nn, activation=activation, name=name + "post_2_")
    return layers.Activation("linear", name=name + "output")(nn)  # Identity, Just need a name here


def MobileViT(
    num_blocks=[1, 3, 3, 5, 4],
    out_channels=[32, 64, 96, 128, 160],
    attn_channels=[0, 0, 144, 192, 240],  # Can be a list matching out_channels, or a float number for expansion ratio of out_channels
    block_types=["conv", "conv", "transform", "transform", "transform"],
    strides=[1, 2, 2, 2, 2],
    expand_ratio=4,
    stem_width=16,
    patch_size=2,
    resize_first=False,  # False for V1, True for V2
    use_depthwise=False,  # False for V1, True for V2
    use_fusion=True,  # True for V1, False for V2
    num_norm_groups=-1,  # -1 or 0 for V1 using layer_norm, or 1 for V2 using group_norm
    use_linear_attention=False,  # False for V1, True for V2
    output_num_features=640,
    layer_scale=-1,
    input_shape=(256, 256, 3),
    num_classes=1000,
    activation="swish",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="mobilevit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = conv2d_no_bias(inputs, stem_width, kernel_size=3, strides=2, padding="same", name="stem_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_")
    height_axis, width_axis, channel_axis = (1, 2, -1) if image_data_format() == "channels_last" else (2, 3, 1)

    # Save line width
    mhsa_mlp_block_common_kwargs = {
        "num_heads": 4,
        "qkv_bias": True,
        "mlp_ratio": 2,
        "num_norm_groups": num_norm_groups,
        "activation": activation,
    }

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    post_activation = activation if use_fusion else None
    for id, (num_block, out_channel, block_type, stride) in enumerate(zip(num_blocks, out_channels, block_types, strides)):
        stack_name = "stack{}_".format(id + 1)
        is_conv_block = True if block_type[0].lower() == "c" else False
        attn_channel = attn_channels[id] if isinstance(attn_channels, (list, tuple)) else make_divisible(attn_channels * out_channel, divisor=8)
        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            stride = stride if block_id == 0 else 1
            use_shortcut = False if stride != 1 or nn.shape[channel_axis] != out_channel else True
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            global_block_id += 1
            if is_conv_block or block_id == 0:  # First transformer block is also a conv block .
                nn = bottle_in_linear_out_block(nn, out_channel, stride, expand_ratio, use_shortcut, block_drop_rate, activation=activation, name=name)
            else:
                if block_id == 1:  # pre
                    pre_attn = nn if use_fusion else None
                    patches_to_batch = not use_linear_attention
                    patch_height = -1 if patches_to_batch else int(math.ceil(nn.shape[height_axis] / patch_size))
                    nn = transformer_pre_process(nn, attn_channel, patch_size, resize_first, use_depthwise, patches_to_batch, activation=activation, name=name)

                if use_linear_attention:  # channels_last for Tensorflow, channels_first for PyTorch
                    nn = linear_mhsa_mlp_block(nn, attn_channel, layer_scale=layer_scale, **mhsa_mlp_block_common_kwargs, name=name)
                else:  # channels_last for both Tensorflow or PyTorch
                    if block_id == 1:
                        block_height, block_width = nn.shape[1:-1]
                        channel_axis = -1
                        nn = functional.reshape(nn, [-1, block_height * block_width, nn.shape[-1]])  # Using 3D for attention inputs
                    nn = mhsa_mlp_block(nn, attn_channel, layer_scale=layer_scale, **mhsa_mlp_block_common_kwargs, name=name)
                    if block_id == num_block - 1:
                        channel_axis = -1 if image_data_format() == "channels_last" else 1
                        nn = functional.reshape(nn, [-1, block_height, block_width, nn.shape[-1]])  # Revert 3D to 4D

                if block_id == num_block - 1:  # post
                    norm_axis = "auto" if use_linear_attention else -1
                    if use_linear_attention:
                        nn = group_norm(nn, groups=num_norm_groups, axis=norm_axis, name=name + "post_")
                    else:
                        nn = layer_norm(nn, axis=norm_axis, name=name + "post_")
                    nn = transformer_post_process(nn, pre_attn, out_channel, patch_size, patch_height, activation=post_activation, name=name)

    nn = output_block(nn, output_num_features, activation, num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="raw01")
    reload_model_weights(model, PRETRAINED_DICT, "mobilevit", pretrained)
    return model


@register_model
def MobileViT_XXS(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [1, 3, 3, 5, 4]
    out_channels = [16, 24, 48, 64, 80]
    attn_channels = [0, 0, 64, 80, 96]
    output_num_features = 320
    expand_ratio = 2
    return MobileViT(**locals(), model_name="mobilevit_xxs", **kwargs)


@register_model
def MobileViT_XS(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [1, 3, 3, 5, 4]
    out_channels = [32, 48, 64, 80, 96]
    attn_channels = 1.5
    output_num_features = 384
    return MobileViT(**locals(), model_name="mobilevit_xs", **kwargs)


@register_model
def MobileViT_S(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [1, 3, 3, 5, 4]
    out_channels = [32, 64, 96, 128, 160]
    attn_channels = 1.5
    return MobileViT(**locals(), model_name="mobilevit_s", **kwargs)



from keras_cv_attention_models.mobilenetv3_family.fbnetv3 import FBNetV3, FBNetV3B, FBNetV3D, FBNetV3G
from keras_cv_attention_models.mobilenetv3_family.lcnet import LCNet, LCNet050, LCNet075, LCNet100, LCNet150, LCNet200, LCNet250
from keras_cv_attention_models.mobilenetv3_family.mobilenetv3 import (
    MobileNetV3,
    MobileNetV3Large,
    MobileNetV3Small,
    MobileNetV3Large075,
    MobileNetV3Large100,
    MobileNetV3Small050,
    MobileNetV3Small075,
    MobileNetV3Small100,
)
from keras_cv_attention_models.mobilenetv3_family.tinynet import TinyNet, TinyNetA, TinyNetB, TinyNetC, TinyNetD, TinyNetE

__head_doc__ = """
Args:
  [Stack parameters]
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  expands: should be a `number` or `list`, indicates hidden expand ratio for each stack.
      Each element can also be a `number` or `list`, indicates hidden expand ratio for each block.
  kernel_sizes: a list matching num_blocks, indicates kernel_size for each stack.
  strides: a list value matching num_blocks, indicates stride in first block of each stack.
  activations: a string for activation of all stacks, or a list value matching num_blocks indicates activation of each stack.
  disable_shortcut: boolean value force disable using shortcut in `inverted_residual_block` for all stacks. True for LCNet, False for others.
  use_blocks_output_activation: boolean value if adding additional activation layer after each block.
      True for LCNet, False for others.
  width_ratio: out_channels expansion ratio.

  [Stem parameters]
  stem_width: output dimension for stem block.
  fix_stem: boolean value if fix stem_width, False for expand with `width_ratio`.
  stem_feature_activation: activation for stem and output feature. "swish" for TinyNet, "hard_swish" for others.

  [SE module parameters]
  se_ratios: a float number for se_ratio of all stacks, or a list value matching num_blocks indicates se_ratio of each stack.
  se_activation: activation for semodule, can be one string value like "relu" for activation after se reduction,
      or two string value like ("relu", "hard_sigmoid_torch") for both activation after se reduction and expansion.
      Set None for se reduction activation same with block activation, expansion activation using "sigmoid".
      ("hard_swish", "hard_sigmoid_torch") for FBNetV3, None for TinyNet, ("relu", "hard_sigmoid_torch") for LCNet and MobileNetV3.
  se_limit_round_down: make sure that round down does not go down by more than [num] reatio in se module.
      0.95 for FBNetV3, 0.9 for others.
  se_divisor: divisor in se module. 1 for TinyNet, 8 for others.
  use_expanded_se_ratio: boolean value contols se reduction ratio.
      False for FBNetV3 and TinyNet, se reduction channel is calculated from `input_channel`, means not expanded with block `expand`.
      True for LCNet and MobileNetV3, se reduction channel is calculated from `hidden_channel`, means also expanded with block `expand`.

  [Output parameters]
  output_num_features: none `0` value to add another `conv2d + bn + activation` layers before classifier output.
  use_additional_output_conv: boolean value if using additional `conv + bn + activation` block after all stacks.
      False for LCNet and TinyNet, True for FBNetV3 and MobileNetV3.
  use_output_feature_bias: boolean value if using bias for `features` block. False for FBNetV3 and TinyNet, True for LCNet and MobileNetV3.
  use_avg_pool_conv_output: boolean value if using `avg_pool -> conv -> activation` or `conv -> bn -> activation -> avg_pool`
      as `features` block. False for TinyNet, True for others.
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`. Set `(None, None, 3)` for dynamic input shape.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  dropout: dropout rate if top layers is included.
  pretrained: None or one of ["imagenet", "miil", "miil_21k", "ssld"].
      ["miil", "miil_21k"] for `MobileNetV3Large100`.
      "ssld" for `LCNet050` `LCNet100` `LCNet250`.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

__mobilenetv3_head_doc__ = """
Keras implementation of [timm/mobilenetv3](https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/mobilenetv3.py).
Paper [PDF 1905.02244 Searching for MobileNetV3](https://arxiv.org/pdf/1905.02244.pdf).
"""

MobileNetV3.__doc__ = __mobilenetv3_head_doc__ + __head_doc__ + """
  [Model common parameters]
""" + __tail_doc__ + """
Model architectures:
  | Model               | Params | FLOPs   | Input | Top1 Acc |
  | ------------------- | ------ | ------- | ----- | -------- |
  | MobileNetV3Small050 | 1.29M  | 24.92M  | 224   | 57.89    |
  | MobileNetV3Small075 | 2.04M  | 44.35M  | 224   | 65.24    |
  | MobileNetV3Small100 | 2.54M  | 57.62M  | 224   | 67.66    |
  | MobileNetV3Large075 | 3.99M  | 156.30M | 224   | 73.44    |
  | MobileNetV3Large100 | 5.48M  | 218.73M | 224   | 75.77    |
  | - miil              | 5.48M  | 218.73M | 224   | 77.92    |
"""

MobileNetV3Large.__doc__ = MobileNetV3.__doc__
MobileNetV3Small.__doc__ = MobileNetV3.__doc__

MobileNetV3Large075.__doc__ = __mobilenetv3_head_doc__ + """
Args:
""" + __tail_doc__

MobileNetV3Large100.__doc__ = MobileNetV3Large075.__doc__
MobileNetV3Small050.__doc__ = MobileNetV3Large075.__doc__
MobileNetV3Small075.__doc__ = MobileNetV3Large075.__doc__
MobileNetV3Small100.__doc__ = MobileNetV3Large075.__doc__


__fbnet_head_doc__ = """
Keras implementation of [timm/mobilenetv3](https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/mobilenetv3.py).
Paper [PDF 2006.02049 FBNetV3: Joint Architecture-Recipe Search using Predictor Pretraining](https://arxiv.org/pdf/2006.02049.pdf).
"""

FBNetV3.__doc__ = __fbnet_head_doc__ + __head_doc__ + """
  [Model common parameters]
""" + __tail_doc__ + """
Model architectures:
  | Model    | Params | FLOPs    | Input | Top1 Acc |
  | -------- | ------ | -------- | ----- | -------- |
  | FBNetV3B | 5.57M  | 539.82M  | 256   | 79.15    |
  | FBNetV3D | 10.31M | 665.02M  | 256   | 79.68    |
  | FBNetV3G | 16.62M | 1379.30M | 256   | 82.05    |
"""

FBNetV3B.__doc__ = __fbnet_head_doc__ + """
Args:
""" + __tail_doc__

FBNetV3D.__doc__ = FBNetV3B.__doc__
FBNetV3G.__doc__ = FBNetV3B.__doc__


__lcnet_head_doc__ = """
Keras implementation of [Github PaddlePaddle/PaddleClas](https://github.com/PaddlePaddle/PaddleClas).
Paper [PDF 2109.15099 PP-LCNet: A Lightweight CPU Convolutional Neural Network](https://arxiv.org/pdf/2109.15099.pdf).
"""

LCNet.__doc__ = __lcnet_head_doc__ + __head_doc__ + """
  [Model common parameters]
""" + __tail_doc__ + """
Model architectures:
  | Model    | Params | FLOPs   | Input | Top1 Acc |
  | -------- | ------ | ------- | ----- | -------- |
  | LCNet050 | 1.88M  | 46.02M  | 224   | 63.10    |
  | - ssld   | 1.88M  | 46.02M  | 224   | 66.10    |
  | LCNet075 | 2.36M  | 96.82M  | 224   | 68.82    |
  | LCNet100 | 2.95M  | 158.28M | 224   | 72.10    |
  | - ssld   | 2.95M  | 158.28M | 224   | 74.39    |
  | LCNet150 | 4.52M  | 338.05M | 224   | 73.71    |
  | LCNet200 | 6.54M  | 585.35M | 224   | 75.18    |
  | LCNet250 | 9.04M  | 900.16M | 224   | 76.60    |
  | - ssld   | 9.04M  | 900.16M | 224   | 80.82    |
"""

LCNet050.__doc__ = __lcnet_head_doc__ + """
Args:
""" + __tail_doc__

LCNet075.__doc__ = LCNet050.__doc__
LCNet100.__doc__ = LCNet050.__doc__
LCNet150.__doc__ = LCNet050.__doc__
LCNet200.__doc__ = LCNet050.__doc__
LCNet250.__doc__ = LCNet050.__doc__


__tinynet_head_doc__ = """
Keras implementation of [Github huawei-noah/CV-Backbones/tinynet_pytorch](https://github.com/huawei-noah/CV-Backbones/tree/master/tinynet_pytorch).
Paper [PDF 2010.14819 Model Rubik’s Cube: Twisting Resolution, Depth and Width for TinyNets](https://arxiv.org/pdf/2010.14819.pdf).
"""

TinyNet.__doc__ = __tinynet_head_doc__ + __head_doc__ + """
  [Model common parameters]
""" + __tail_doc__ + """
Model architectures:
  | Model    | Params | FLOPs   | Input | Top1 Acc |
  | -------- | ------ | ------- | ----- | -------- |
  | TinyNetE | 2.04M  | 25.22M  | 106   | 59.86    |
  | TinyNetD | 2.34M  | 53.35M  | 152   | 66.96    |
  | TinyNetC | 2.46M  | 103.22M | 184   | 71.23    |
  | TinyNetB | 3.73M  | 206.28M | 188   | 74.98    |
  | TinyNetA | 6.19M  | 343.74M | 192   | 77.65    |
"""

TinyNetE.__doc__ = __tinynet_head_doc__ + """
Args:
""" + __tail_doc__

TinyNetD.__doc__ = TinyNetE.__doc__
TinyNetC.__doc__ = TinyNetE.__doc__
TinyNetB.__doc__ = TinyNetE.__doc__
TinyNetA.__doc__ = TinyNetE.__doc__



import os
import tensorflow as tf
from tensorflow import keras


def init_mean_std_by_rescale_mode(rescale_mode):
    if isinstance(rescale_mode, (list, tuple)):  # Specific mean and std
        mean, std = rescale_mode
    elif rescale_mode == "torch":
        mean = tf.constant([0.485, 0.456, 0.406]) * 255.0
        std = tf.constant([0.229, 0.224, 0.225]) * 255.0
    elif rescale_mode == "tf":  # [0, 255] -> [-1, 1]
        mean, std = 127.5, 127.5
        # mean, std = 127.5, 128.0
    elif rescale_mode == "tf128":  # [0, 255] -> [-1, 1]
        mean, std = 128.0, 128.0
    elif rescale_mode == "raw01":
        mean, std = 0, 255.0  # [0, 255] -> [0, 1]
    elif rescale_mode == "clip":  # value from openai/CLIP
        mean = tf.constant([0.48145466, 0.4578275, 0.40821073]) * 255.0
        std = tf.constant([0.26862954, 0.26130258, 0.27577711]) * 255.0
    else:
        mean, std = 0, 1  # raw inputs [0, 255]
    return mean, std


def tf_imread(file_path):
    # tf.print('Reading file:', file_path)
    img = tf.io.read_file(file_path)
    # img = tf.image.decode_jpeg(img, channels=3)  # [0, 255]
    img = tf.image.decode_image(img, channels=3, expand_animations=False)  # [0, 255]
    img = tf.cast(img, "float32")  # [0, 255]
    return img


def random_crop_fraction(size, scale=(0.08, 1.0), ratio=(0.75, 1.3333333), log_distribute=True, compute_dtype="float32"):
    height, width = tf.cast(size[0], dtype=compute_dtype), tf.cast(size[1], dtype=compute_dtype)
    area = height * width
    scale_max = tf.minimum(tf.minimum(height * height * ratio[1] / area, width * width / ratio[0] / area), scale[1])
    target_area = tf.random.uniform((), scale[0], scale_max, dtype=compute_dtype) * area

    ratio_min = tf.maximum(target_area / (height * height), ratio[0])
    ratio_max = tf.minimum(width * width / target_area, ratio[1])
    if log_distribute:  # More likely to select a smaller value
        log_min, log_max = tf.math.log(ratio_min), tf.math.log(ratio_max)
        aspect_ratio = tf.random.uniform((), log_min, log_max, dtype=compute_dtype)
        aspect_ratio = tf.math.exp(aspect_ratio)
    else:
        aspect_ratio = tf.random.uniform((), ratio_min, ratio_max, dtype=compute_dtype)

    ww_crop = tf.cast(tf.math.floor(tf.sqrt(target_area * aspect_ratio)), "int32")
    hh_crop = tf.cast(tf.math.floor(tf.sqrt(target_area / aspect_ratio)), "int32")
    # tf.print(">>>> height, width, hh_crop, ww_crop, hh_fraction_min, hh_fraction_max:", height, width, hh_crop, ww_crop, hh_fraction_min, hh_fraction_max)
    # return hh_crop, ww_crop, target_area, hh_fraction_min, hh_fraction_max, hh_fraction
    return hh_crop, ww_crop
    # return hh_fraction, target_area / hh_fraction # float value will stay in scale and ratio range exactly


def random_crop_and_resize_image(image, target_shape, scale=(0.08, 1.0), ratio=(0.75, 1.3333333), method="bilinear", antialias=False):
    """Random crop and resize, return `image, scale_hh, scale_ww, crop_hh, crop_ww`"""
    height, width = tf.shape(image)[0], tf.shape(image)[1]
    cropped_hh, cropped_ww = random_crop_fraction((height, width), scale, ratio)
    cropped_hh, cropped_ww = tf.clip_by_value(cropped_hh, 1, height - 1), tf.clip_by_value(cropped_ww, 1, width - 1)
    crop_hh = tf.random.uniform((), 0, height - cropped_hh, dtype=cropped_hh.dtype)
    crop_ww = tf.random.uniform((), 0, width - cropped_ww, dtype=cropped_ww.dtype)
    image = image[crop_hh : crop_hh + cropped_hh, crop_ww : crop_ww + cropped_ww]
    image = tf.image.resize(image, target_shape, method=method, antialias=antialias)

    # crop_hh, crop_ww is crop size after image size applyied scale.
    # image size firstly rescale with (scale_hh, scale_ww), then crop as [crop_hh: crop_hh + target_shape, crop_ww: crop_ww + target_shape]
    # scale_hh = tf.cast(target_shape[0], "float32") / tf.cast(cropped_hh, "float32")
    # scale_ww = tf.cast(target_shape[1], "float32") / tf.cast(cropped_ww, "float32")
    # crop_hh = tf.cast(tf.cast(crop_hh, "float32") * scale_hh, "int32")
    # crop_ww = tf.cast(tf.cast(crop_ww, "float32") * scale_ww, "int32")

    # For coords or bbox value in (0, 1) -> * scale - offset
    # (bbox * height * scale_hh - crop_hh) / target_shape[0]
    scale_hh = tf.cast(height, "float32") / tf.cast(cropped_hh, "float32")
    scale_ww = tf.cast(width, "float32") / tf.cast(cropped_ww, "float32")
    crop_hh = tf.cast(crop_hh, "float32") / tf.cast(cropped_hh, "float32")
    crop_ww = tf.cast(crop_ww, "float32") / tf.cast(cropped_ww, "float32")

    return image, scale_hh, scale_ww, crop_hh, crop_ww


def random_erasing_per_pixel(image, num_layers=1, scale=(0.02, 0.33333333), ratio=(0.3, 3.3333333), probability=0.5):
    """https://github.com/rwightman/pytorch-image-models/blob/master/timm/data/random_erasing.py"""
    if tf.random.uniform(()) > probability:
        return image

    mean = [123.675, 116.28, 103.53]
    std = [58.395, 57.120003, 57.375]
    height, width, _ = image.shape
    for _ in range(num_layers):
        hh, ww = random_crop_fraction((height, width), scale=scale, ratio=ratio)
        hh_ss = tf.random.uniform((), 0, height - hh, dtype="int32")
        ww_ss = tf.random.uniform((), 0, width - ww, dtype="int32")
        mask = tf.random.normal([hh, ww, 3], mean=mean, stddev=std)
        mask = tf.clip_by_value(mask, 0.0, 255.0)  # value in [0, 255]
        aa = tf.concat([image[:hh_ss, ww_ss : ww_ss + ww], mask, image[hh_ss + hh :, ww_ss : ww_ss + ww]], axis=0)
        image = tf.concat([image[:, :ww_ss], aa, image[:, ww_ss + ww :]], axis=1)
    return image


def sample_beta_distribution(shape, concentration_0=0.4, concentration_1=0.4):
    gamma_1_sample = tf.random.gamma(shape=shape, alpha=concentration_1)
    gamma_2_sample = tf.random.gamma(shape=shape, alpha=concentration_0)
    return gamma_1_sample / (gamma_1_sample + gamma_2_sample)


def mixup(images, labels, token_label=None, alpha=0.4, min_mix_weight=0):
    # mix_weight = tfp.distributions.Beta(alpha, alpha).sample([batch_size, 1])
    batch_size = tf.shape(images)[0]
    mix_weight = sample_beta_distribution([batch_size], alpha, alpha)
    mix_weight = tf.maximum(mix_weight, 1.0 - mix_weight)

    # For min_mix_weight=0.1, regard values with `> 0.9` as no mixup, this probability is near `1 - alpha`
    # alpha: no_mixup --> {0.1: 0.8128, 0.2: 0.6736, 0.4: 0.4793, 0.6: 0.3521, 0.8: 0.2636, 1.0: 0.2000}
    if min_mix_weight > 0:
        mix_weight = tf.where(mix_weight > 1 - min_mix_weight, tf.ones_like(mix_weight), mix_weight)

    label_mix_weight = tf.cast(tf.expand_dims(mix_weight, -1), "float32")
    img_mix_weight = tf.cast(tf.reshape(mix_weight, [batch_size, 1, 1, 1]), images.dtype)

    labels = tf.cast(labels, "float32")
    # images = images * img_mix_weight + images[::-1] * (1.0 - img_mix_weight)
    # labels = labels * label_mix_weight + labels[::-1] * (1 - label_mix_weight)
    shuffle_index = tf.random.shuffle(tf.range(batch_size))
    images = images * img_mix_weight + tf.gather(images, shuffle_index) * (1.0 - img_mix_weight)
    labels = labels * label_mix_weight + tf.gather(labels, shuffle_index) * (1 - label_mix_weight)
    if token_label is None:
        return images, labels
    else:
        # token_label shape `[batch, path_height, patch_width, one_hot_labels]`
        token_label = token_label * img_mix_weight + tf.gather(token_label, shuffle_index) * (1 - img_mix_weight)
        return images, labels, token_label


def get_box(mix_weight, height, width, dtype="int32"):
    cut_rate_half = tf.math.sqrt(1.0 - mix_weight) / 2
    cut_h_half, cut_w_half = tf.cast(cut_rate_half * float(height), dtype), tf.cast(cut_rate_half * float(width), dtype)
    cut_h_half, cut_w_half = tf.maximum(cut_h_half, 1), tf.maximum(cut_w_half, 1)
    # center_y = tf.random.uniform((), minval=cut_h_half, maxval=height - cut_h_half, dtype=dtype)
    # center_x = tf.random.uniform((), minval=cut_w_half, maxval=width - cut_w_half, dtype=dtype)
    # return center_y - cut_h_half, center_x - cut_w_half, cut_h_half * 2, cut_w_half * 2

    # Can be non-square on border
    center_y = tf.random.uniform((), minval=0, maxval=height, dtype=dtype)
    center_x = tf.random.uniform((), minval=0, maxval=width, dtype=dtype)
    yl = tf.clip_by_value(center_y - cut_h_half, 0, height)
    yr = tf.clip_by_value(center_y + cut_h_half, 0, height)
    xl = tf.clip_by_value(center_x - cut_w_half, 0, width)
    xr = tf.clip_by_value(center_x + cut_w_half, 0, width)
    return yl, xl, yr - yl, xr - xl


def cutmix(images, labels, token_label=None, alpha=0.5, min_mix_weight=0):
    # Get a sample from the Beta distribution
    batch_size = tf.shape(images)[0]
    _, hh, ww, _ = images.shape
    mix_weight = sample_beta_distribution((), alpha, alpha)  # same value in batch
    if token_label is None:
        offset_height, offset_width, target_height, target_width = get_box(mix_weight, hh, ww)
        mix_weight = 1.0 - tf.cast(target_height * target_width, "float32") / tf.cast(hh * ww, "float32")
    else:
        # token_label shape `[batch, path_height, patch_width, one_hot_labels]`
        # Limit box within patchs
        _, path_height_int, patch_width_int, _ = token_label.shape
        path_height, patch_width = float(path_height_int), float(patch_width_int)
        tl_offset_height, tl_offset_width, tl_target_height, tl_target_width = get_box(mix_weight, path_height, patch_width, dtype="float32")
        offset_height, offset_width = tf.cast(tl_offset_height / path_height * float(hh), "int32"), tf.cast(tl_target_width / patch_width * float(ww), "int32")
        target_height, target_width = tf.cast(tl_target_height / path_height * float(hh), "int32"), tf.cast(tl_target_width / patch_width * float(ww), "int32")
        tl_offset_height, tl_offset_width = tf.cast(tl_offset_height, "int32"), tf.cast(tl_offset_width, "int32")
        tl_target_height, tl_target_width = tf.cast(tl_target_height, "int32"), tf.cast(tl_target_width, "int32")
        target_height, target_width = tf.clip_by_value(target_height, 0, hh - offset_height), tf.clip_by_value(target_width, 0, ww - offset_width)

    if mix_weight < min_mix_weight or 1 - mix_weight < min_mix_weight:
        # For input_shape=224, min_mix_weight=0.1, min_height = 224 * sqrt(0.1) = 70.835
        return (images, labels) if token_label is None else (images, labels, token_label)

    crops = tf.image.crop_to_bounding_box(images, offset_height, offset_width, target_height, target_width)
    pad_crops = tf.image.pad_to_bounding_box(crops, offset_height, offset_width, hh, ww)

    labels = tf.cast(labels, "float32")
    # images = images - pad_crops + pad_crops[::-1]
    # labels = labels * mix_weight + labels[::-1] * (1.0 - mix_weight)
    shuffle_index = tf.random.shuffle(tf.range(batch_size))
    images = images - pad_crops + tf.gather(pad_crops, shuffle_index)
    labels = labels * mix_weight + tf.gather(labels, shuffle_index) * (1.0 - mix_weight)
    if token_label is None:
        return images, labels
    else:
        # token_label shape `[batch, path_height, patch_width, one_hot_labels]`
        # tf.print((path_height_int, patch_width_int), (tl_offset_height, tl_offset_width), (tl_target_height, tl_target_width))
        token_label_crops = tf.image.crop_to_bounding_box(token_label, tl_offset_height, tl_offset_width, tl_target_height, tl_target_width)
        token_label_pad_crops = tf.image.pad_to_bounding_box(token_label_crops, tl_offset_height, tl_offset_width, path_height_int, patch_width_int)
        token_label = token_label - token_label_pad_crops + tf.gather(token_label_pad_crops, shuffle_index)
        # token_label = (token_label - token_label_pad_crops) * mix_weight + (tf.gather(token_label_pad_crops, shuffle_index)) * (1.0 - mix_weight)
        return images, labels, token_label


def apply_mixup_cutmix(train_dataset, mixup_alpha, cutmix_alpha, switch_prob=0.5):
    if mixup_alpha > 0 and mixup_alpha <= 1 and cutmix_alpha > 0 and cutmix_alpha <= 1:
        print(">>>> Both mixup_alpha and cutmix_alpha provided: mixup_alpha = {}, cutmix_alpha = {}".format(mixup_alpha, cutmix_alpha))
        mix_func = lambda *args: tf.cond(
            tf.random.uniform(()) > switch_prob,  # switch_prob = 0.5
            lambda: mixup(*args, alpha=mixup_alpha),
            lambda: cutmix(*args, alpha=cutmix_alpha),
        )
    elif mixup_alpha > 0 and mixup_alpha <= 1:
        print(">>>> mixup_alpha provided:", mixup_alpha)
        mix_func = lambda *args: mixup(*args, alpha=mixup_alpha)
    elif cutmix_alpha > 0 and cutmix_alpha <= 1:
        print(">>>> cutmix_alpha provided:", cutmix_alpha)
        mix_func = lambda *args: cutmix(*args, alpha=cutmix_alpha)
    else:
        return train_dataset
    return train_dataset.map(mix_func, num_parallel_calls=tf.data.AUTOTUNE)


class RandomProcessImage:
    def __init__(
        self,
        target_shape=(224, 224),
        central_crop=1.0,
        random_crop_min=1.0,
        resize_method="bilinear",
        resize_antialias=False,
        random_erasing_prob=0.0,
        random_erasing_layers=1,
        magnitude=0,
        num_layers=2,
        use_cutout=False,
        use_relative_translate=True,
        use_color_increasing=True,
        use_positional_related_ops=True,
        use_token_label=False,
        token_label_target_patches=-1,
        num_classes=1000,
        **randaug_kwargs,
    ):
        self.magnitude, self.random_erasing_prob, self.use_token_label = magnitude, random_erasing_prob, use_token_label
        self.target_shape = target_shape if len(target_shape) == 2 else target_shape[:2]
        self.central_crop, self.random_crop_min = central_crop, random_crop_min
        self.resize_method, self.resize_antialias = resize_method, resize_antialias

        if random_erasing_prob > 0:
            self.random_erasing = lambda img: random_erasing_per_pixel(img, num_layers=random_erasing_layers, probability=random_erasing_prob)
            use_cutout = False

        if magnitude > 0:
            from keras_cv_attention_models.imagenet import augment

            # for target_shape = 224, translate_const = 100 and cutout_const = 40
            translate_const = 0.45 if use_relative_translate else min(self.target_shape) * 0.45
            cutout_const = min(self.target_shape) * 0.18
            print(">>>> RandAugment: magnitude = %d, translate_const = %f, cutout_const = %f" % (magnitude, translate_const, cutout_const))

            self.randaug = augment.RandAugment(
                num_layers=num_layers,
                magnitude=magnitude,
                translate_const=translate_const,
                cutout_const=cutout_const,
                use_cutout=use_cutout,
                use_relative_translate=use_relative_translate,
                use_color_increasing=use_color_increasing,
                use_positional_related_ops=use_positional_related_ops,
                **randaug_kwargs,
            )

        if use_token_label:
            from keras_cv_attention_models.imagenet import token_label

            self.token_label_align = token_label.TokenLabelAlign(num_classes=num_classes, target_num_pathes=token_label_target_patches)

    def __call__(self, image, token_label=None):
        # image = datapoint["image"]
        if len(image.shape) < 2:
            image = tf_imread(image)
        channel = image.shape[-1]

        flip_left_right, scale_hh, scale_ww, crop_hh, crop_ww = tf.cast(False, tf.bool), 1, 1, 0, 0  # Init value
        if self.random_crop_min > 0 and self.random_crop_min < 1:
            image, scale_hh, scale_ww, crop_hh, crop_ww = random_crop_and_resize_image(
                image, self.target_shape, scale=(self.random_crop_min, 1.0), method=self.resize_method, antialias=self.resize_antialias
            )
        # elif self.central_crop > 0:
        #     image = tf.image.central_crop(image, self.central_crop)
        else:
            image = tf.image.resize(image, self.target_shape, method=self.resize_method, antialias=self.resize_antialias)

        if self.magnitude >= 0:
            # tf.image.random_flip_left_right
            flip_left_right = tf.random.uniform(()) < 0.5
            image = tf.cond(flip_left_right, lambda: tf.image.flip_left_right(image), lambda: image)
        if self.magnitude > 0:
            image = self.randaug(image)
        if self.random_erasing_prob > 0:
            image = self.random_erasing(image)

        image = tf.cast(image, tf.float32)
        image.set_shape([*self.target_shape[:2], channel])

        # label = datapoint["label"]
        if self.use_token_label and token_label is not None:
            token_label = self.token_label_align(token_label, flip_left_right, scale_hh, scale_ww, crop_hh, crop_ww)
            return image, token_label
        else:
            return image


def evaluation_process_crop_resize(image, target_shape=(224, 224), central_crop=1.0, resize_method="bilinear", antialias=False):
    # image = datapoint["image"]
    if len(image.shape) < 3:
        image = tf_imread(image)
    if central_crop > 0:  # Do not crop if central_crop == -1
        shape = tf.shape(image)
        height, width = shape[0], shape[1]
        crop_size = tf.cast((central_crop * tf.cast(tf.minimum(height, width), tf.float32)), tf.int32)
        y, x = (height - crop_size) // 2, (width - crop_size) // 2
        image = tf.image.crop_to_bounding_box(image, y, x, crop_size, crop_size)
    image = tf.image.resize(image, target_shape, method=resize_method, antialias=antialias)
    return image


# Not using
def evaluation_process_resize_crop(image, target_shape=(224, 224), central_crop=1.0, resize_method="bilinear", antialias=False):
    # image = datapoint["image"]
    if len(image.shape) < 3:
        image = tf_imread(image)
    shape = tf.shape(image)
    height, width = shape[0], shape[1]
    min_border = tf.cast(tf.minimum(height, width), tf.float32)
    scale_size = tf.cast(tf.minimum(*target_shape), tf.float32) / central_crop
    hh_scale = tf.cast(tf.floor(tf.cast(height, tf.float32) * scale_size / min_border), tf.int32)
    ww_scale = tf.cast(tf.floor(tf.cast(width, tf.float32) * scale_size / min_border), tf.int32)
    image = tf.image.resize(image, (hh_scale, ww_scale), method=resize_method, antialias=antialias)

    y, x = (hh_scale - target_shape[0]) // 2, (ww_scale - target_shape[1]) // 2
    image = tf.image.crop_to_bounding_box(image, y, x, target_shape[0], target_shape[1])
    return image


def init_from_json_or_csv_or_tsv(data_path, is_caption=False):
    if data_path.endswith(".json"):
        import json

        with open(data_path, "r") as ff:
            aa = json.load(ff)
        test_key = "validation" if "validation" in aa else "test"
        train, test, info = aa["train"], aa[test_key], aa.get("info", {})
    else:
        import csv

        delimiter = "\t" if data_path.endswith(".tsv") else ","
        label_key = "caption" if is_caption else "label"
        train, test, info, is_train = [], [], {}, True
        with open(data_path) as ff:
            for ii in csv.reader(ff, delimiter=delimiter):
                if ii[0] in ["base_path", "num_classes"]:  # special keys for info
                    info[ii[0]] = ii[1]
                    continue

                if ii[0] == "TEST":  # Use this as indicator for start of test set
                    is_train = False
                elif is_train:
                    train.append({"image": ii[0], label_key: ii[1]})
                else:
                    test.append({"image": ii[0], label_key: ii[1]})
        test_key = "test"

    """ Construct more info """
    total_images, num_classes = len(train), info.get("num_classes", 0)
    if not is_caption and num_classes <= 0 and "label" in train[0] and isinstance(train[0]["label"], int):
        num_classes = max([ii["label"] for ii in train]) + 1
        print(">>>> Using max value from train as num_classes:", num_classes)

    if "base_path" in info and len(info["base_path"]) > 0:
        base_path = os.path.expanduser(info["base_path"])
        for ii in train:
            ii["image"] = os.path.join(base_path, ii["image"])
        for ii in test:
            ii["image"] = os.path.join(base_path, ii["image"])
    num_channels = tf_imread(train[0]["image"]).shape[-1]
    return (train, test, info, test_key), (total_images, num_classes, num_channels)


def build_custom_dataset(data_path, with_info=False, info_only=False, caption_tokenizer=None):
    is_caption = False if caption_tokenizer is None else True
    (train, test, info, test_key), (total_images, num_classes, num_channels) = init_from_json_or_csv_or_tsv(data_path, is_caption)

    if is_caption:
        # from tqdm import tqdm
        # train = [{"image": ii["image"], "caption": caption_tokenizer(ii["caption"])} for ii in tqdm(train, "Tokenizing train caption")]
        # test = [{"image": ii["image"], "caption": caption_tokenizer(ii["caption"])} for ii in tqdm(test, "Tokenizing test caption")]
        context_length = caption_tokenizer.context_length
        output_signature = {"image": tf.TensorSpec(shape=(), dtype=tf.string), "caption": tf.TensorSpec(shape=(context_length,), dtype=tf.int64)}
        train_gen = lambda: ({"image": ii["image"], "caption": caption_tokenizer(ii["caption"])} for ii in train)
        test_gen = lambda: ({"image": ii["image"], "caption": caption_tokenizer(ii["caption"])} for ii in test)
    else:
        output_signature = {"image": tf.TensorSpec(shape=(), dtype=tf.string), "label": tf.TensorSpec(shape=(), dtype=tf.int64)}
        train_gen = lambda: (ii for ii in train)
        test_gen = lambda: (ii for ii in test)
    train_ds = tf.data.Dataset.from_generator(train_gen, output_signature=output_signature)
    test_ds = tf.data.Dataset.from_generator(test_gen, output_signature=output_signature)

    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    train_ds = train_ds.apply(tf.data.experimental.assert_cardinality(len(train))).with_options(options)
    test_ds = test_ds.apply(tf.data.experimental.assert_cardinality(len(test))).with_options(options)
    dataset = {"train": train_ds, test_key: test_ds}
    return (dataset, total_images, num_classes, num_channels) if with_info else dataset


def build_token_label_dataset(train_dataset, token_label_file):
    import pickle

    with open(token_label_file, "rb") as ff:
        token_label_data = pickle.load(ff)
    token_label_ds = tf.data.Dataset.from_tensor_slices(token_label_data)
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    token_label_ds = token_label_ds.with_options(options)

    token_label_train_ds = tf.data.Dataset.zip((train_dataset, token_label_ds))
    return token_label_train_ds


def build_distillation_dataset(ds, teacher_model, input_shape, resize_method="bilinear", resize_antialias=False):
    teacher_model.trainable = False
    # Using teacher_model outputs instead of actual labels
    # Using `ds = ds.map(lambda xx, yy: (xx, teacher_model(xx)))` will run teacher_model on CPU, though will on GPU if XLA enabled

    # print(f">>>> {ds.element_spec[0].shape = }, {input_shape = }")
    # gen_func = lambda: ((tf.image.resize(xx, input_shape[:2], method=resize_method, antialias=resize_antialias), (yy, teacher_model(xx))) for xx, yy in ds)
    # image_signature = tf.TensorSpec(shape=(None, input_shape[0], input_shape[1], ds.element_spec[0].shape[-1]), dtype=tf.float32)
    # output_signature = (image_signature, (ds.element_spec[1], ds.element_spec[1]))
    # new_ds = tf.data.Dataset.from_generator(gen_func, output_signature=output_signature)

    output_signature = (ds.element_spec[0], (ds.element_spec[1], ds.element_spec[1]))
    new_ds = tf.data.Dataset.from_generator(lambda: ((xx, (yy, teacher_model(xx))) for xx, yy in ds), output_signature=output_signature)
    if ds.element_spec[0].shape[1] != input_shape[0] or ds.element_spec[0].shape[2] != input_shape[1]:
        resize_func = lambda xx, yy: (tf.image.resize(xx, input_shape[:2], method=resize_method, antialias=resize_antialias), yy)
        new_ds = new_ds.map(resize_func, num_parallel_calls=tf.data.AUTOTUNE)

    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    new_ds = new_ds.apply(tf.data.experimental.assert_cardinality(len(ds))).with_options(options)
    return new_ds


def init_dataset(
    data_name="imagenet2012",  # dataset params
    input_shape=(224, 224),
    batch_size=64,
    buffer_size=1000,
    info_only=False,
    mixup_alpha=0,  # mixup / cutmix params
    cutmix_alpha=0,
    rescale_mode="tf",  # rescale mode, ["tf", "torch"], or specific `(mean, std)` like `(128.0, 128.0)`
    eval_central_crop=1.0,  # augment params
    random_crop_min=1.0,
    resize_method="bilinear",  # ["bilinear", "bicubic"]
    resize_antialias=False,
    random_erasing_prob=0.0,
    magnitude=0,
    num_layers=2,
    use_positional_related_ops=True,
    use_shuffle=True,
    seed=None,
    token_label_file=None,
    token_label_target_patches=-1,
    teacher_model=None,
    teacher_model_input_shape=-1,  # -1 means same with input_shape
    **augment_kwargs,  # Too many...
):
    import tensorflow_datasets as tfds

    # print(">>>> Dataset args:", locals())
    is_tpu = True if len(tf.config.list_logical_devices("TPU")) > 0 else False  # Set True for try_gcs and drop_remainder
    try_gcs, drop_remainder = is_tpu, is_tpu
    use_token_label = False if token_label_file is None else True
    use_distill = False if teacher_model is None else True
    teacher_model_input_shape = input_shape if teacher_model_input_shape == -1 else teacher_model_input_shape

    if data_name.endswith(".json") or data_name.endswith(".tsv"):
        if info_only:
            _, (total_images, num_classes, num_channels) = init_from_json_or_csv_or_tsv(data_name)
        else:
            dataset, total_images, num_classes, num_channels = build_custom_dataset(data_name, with_info=True)
    else:
        dataset, info = tfds.load(data_name, with_info=True, try_gcs=try_gcs)
        num_classes = info.features["label"].num_classes
        num_channels = info.features["image"].shape[-1]
        total_images = info.splits["train"].num_examples
    steps_per_epoch = int(tf.math.ceil(total_images / float(batch_size)))
    if info_only:
        return total_images, num_classes, steps_per_epoch, num_channels  # return num_channels in case it's not 3

    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    AUTOTUNE = tf.data.AUTOTUNE

    """ Train dataset functions """
    train_image_func = RandomProcessImage(
        target_shape=teacher_model_input_shape if use_distill else input_shape,
        central_crop=-1,  # Resize directly w/o crop, if random_crop_min not in (0, 1)
        random_crop_min=random_crop_min,
        resize_method=resize_method,
        resize_antialias=resize_antialias,
        random_erasing_prob=random_erasing_prob,
        magnitude=magnitude,
        num_layers=num_layers,
        use_positional_related_ops=use_positional_related_ops,
        use_token_label=use_token_label,
        token_label_target_patches=token_label_target_patches,
        num_classes=num_classes,
        **augment_kwargs,
    )

    if use_token_label:
        train_pre_batch = lambda data_point: (*train_image_func(data_point["image"]), data_point["label"])
        train_post_batch = lambda xx, token_label, yy: ((xx - mean) / std, tf.one_hot(yy, num_classes), token_label)
    else:
        train_pre_batch = lambda data_point: (train_image_func(data_point["image"]), data_point["label"])
        train_post_batch = lambda xx, yy: ((xx - mean) / std, tf.one_hot(yy, num_classes))

    """ Train dataset """
    train_dataset = dataset["train"]
    if use_token_label:
        train_dataset = build_token_label_dataset(train_dataset, token_label_file)
    if use_shuffle:
        train_dataset = train_dataset.shuffle(buffer_size, seed=seed)
    train_dataset = train_dataset.map(train_pre_batch, num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.batch(batch_size, drop_remainder=drop_remainder).map(train_post_batch, num_parallel_calls=AUTOTUNE)
    train_dataset = apply_mixup_cutmix(train_dataset, mixup_alpha, cutmix_alpha, switch_prob=0.5)

    if use_token_label:
        train_dataset = train_dataset.map(lambda xx, yy, token_label: (xx, (yy, token_label)))
    elif use_distill:
        print(">>>> KLDivergence teacher model provided.")
        train_dataset = build_distillation_dataset(train_dataset, teacher_model, input_shape)

    train_dataset = train_dataset.prefetch(buffer_size=AUTOTUNE)
    # return train_dataset

    """ Test dataset """
    test_dataset = dataset.get("validation", dataset.get("test", None))
    if test_dataset is not None:
        # test_pre_batch = lambda xx: evaluation_process_resize_crop(xx, input_shape[:2], eval_central_crop, resize_method, resize_antialias)  # timm
        # test_image_func = lambda xx: evaluation_process_crop_resize(xx, input_shape[:2], eval_central_crop, resize_method, resize_antialias)
        test_pre_batch = lambda data_point: (
            evaluation_process_crop_resize(data_point["image"], input_shape[:2], eval_central_crop, resize_method, resize_antialias),
            data_point["label"],
        )
        if use_token_label:
            test_post_batch = lambda xx, yy: ((xx - mean) / std, (tf.one_hot(yy, num_classes), None))  # just give None on token_label data position
        elif use_distill:
            test_post_batch = lambda xx, yy: ((xx - mean) / std, (tf.one_hot(yy, num_classes), None))
        else:
            test_post_batch = lambda xx, yy: ((xx - mean) / std, tf.one_hot(yy, num_classes))

        test_dataset = test_dataset.map(test_pre_batch, num_parallel_calls=AUTOTUNE)
        # Have to drop_remainder also for test set...
        test_dataset = test_dataset.batch(batch_size, drop_remainder=drop_remainder).map(test_post_batch)
    return train_dataset, test_dataset, total_images, num_classes, steps_per_epoch


""" Show """


def show_batch_sample(dataset, rescale_mode="tf", rows=-1, caption_tokenizer=None, base_size=3, indices_2_labels=None):
    from keras_cv_attention_models import plot_func
    from keras_cv_attention_models.imagenet.eval_func import decode_predictions

    if isinstance(dataset, (list, tuple)):
        images, labels = dataset
    elif isinstance(dataset.element_spec[0], tuple):  # caption datasets
        (images, labels), _ = dataset.as_numpy_iterator().next()
    elif isinstance(dataset.element_spec[1], tuple):  # token_label datasets
        images, (labels, token_label) = dataset.as_numpy_iterator().next()
    else:
        images, labels = dataset.as_numpy_iterator().next()

    if caption_tokenizer is not None:
        labels = [caption_tokenizer(ii) for ii in labels]

    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    mean, std = (mean.numpy(), std.numpy()) if hasattr(mean, "numpy") else (mean, std)
    images = (images * std + mean) / 255

    if isinstance(labels[0], str):
        pass  # caption datasets
    elif tf.shape(labels)[-1] == 1000:
        labels = [ii[0][1] for ii in decode_predictions(labels, top=1)]
    elif tf.rank(labels[0]) == 1:
        labels = tf.argmax(labels, axis=-1).numpy()  # If 2 dimension

    if not isinstance(labels[0], str) and indices_2_labels is not None:
        labels = [indices_2_labels.get(label, indices_2_labels.get(str(label), str(label))) for label in labels]
    ax, _ = plot_func.stack_and_plot_images(images, texts=labels, rows=rows, ax=None, base_size=base_size)
    return ax


def show_token_label_patches_single(image, token_label, rescale_mode="tf", top_k=3, resize_patch_shape=(160, 160)):
    from keras_cv_attention_models import plot_func

    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    mean, std = (mean.numpy(), std.numpy()) if hasattr(mean, "numpy") else (mean, std)
    image = (image * std + mean) / 255

    height, width = image.shape[:2]
    num_height_patch, num_width_patch = token_label.shape[0], token_label.shape[1]
    height_patch, width_patch = int(tf.math.ceil(height / num_height_patch)), int(tf.math.ceil(width / num_width_patch))
    token_label_scores, token_label_classes = tf.math.top_k(token_label, top_k)
    token_label_scores, token_label_classes = token_label_scores.numpy(), token_label_classes.numpy()
    # fig, axes = plt.subplots(num_height_patch, num_width_patch)

    image_pathes, labels = [], []
    for hh_id in range(num_height_patch):
        hh_image = image[hh_id * height_patch : (hh_id + 1) * height_patch]
        for ww_id in range(num_width_patch):
            image_patch = hh_image[:, ww_id * width_patch : (ww_id + 1) * width_patch]
            image_pathes.append(tf.image.resize(image_patch, resize_patch_shape).numpy())
            scores = ",".join(["{:.1f}".format(ii * 100) for ii in token_label_scores[hh_id, ww_id]])
            classes = ",".join(["{:d}".format(ii) for ii in token_label_classes[hh_id, ww_id].astype("int")])
            labels.append(classes + "\n" + scores)
    plot_func.stack_and_plot_images(image_pathes, labels)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    ChannelAffine,
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    mlp_block,
    multi_head_self_attention,
    se_module,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

BATCH_NORM_EPSILON = 1e-5
PRETRAINED_DICT = {
    "fastvit_ma36": {"imagenet": "13d53c4cfb2587962e779027efdc1104", "distill": "8a0e72c0171b59091b484849f4b7cd87"},
    "fastvit_s12": {"imagenet": "de71db346a180d81e7195291b926b393", "distill": "a0d97a2ef840ab403d12513d40807ffd"},
    "fastvit_sa12": {"imagenet": "4c3747e8915eec9359395984dd832555", "distill": "d656d94844af0e34ec57f38b363395a3"},
    "fastvit_sa24": {"imagenet": "11dd3d252e8ad205a69274d196a7bea2", "distill": "6465e5f6218735eae17a06ca27839a64"},
    "fastvit_sa36": {"imagenet": "b93fed540c0b5acc1cdd99b54173ecc0", "distill": "0ef091d7ceeecfc5a4c2d05d5ebdd297"},
    "fastvit_t12": {"imagenet": "2f6e745fbfdfcd375bd10e8e3d58663f", "distill": "2880385ec6a2225c59852a18b3be0bbb"},
    "fastvit_t8": {"imagenet": "58b3cf9b0f5072b7b0a605c60f76328d", "distill": "90147a9dd53178f5c046c6cdd83137c0"},
}


def rep_conv_block(inputs, out_channel, kernel_size=3, strides=1, groups=1, deploy=False, name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    use_depthwise = input_channel == out_channel and groups == out_channel
    if deploy and use_depthwise:
        return depthwise_conv2d_no_bias(inputs, kernel_size=kernel_size, strides=strides, use_bias=True, padding="same", name=name + "REPARAM_1_")
    elif deploy:
        return conv2d_no_bias(
            inputs, out_channel, kernel_size=kernel_size, strides=strides, use_bias=True, padding="same", groups=groups, name=name + "REPARAM_1_"
        )

    use_shortcut = input_channel == out_channel and strides == 1
    if use_shortcut:
        short = batchnorm_with_activation(inputs, epsilon=BATCH_NORM_EPSILON, name=name + "REPARAM_0_")

    if use_depthwise:
        dw_1 = depthwise_conv2d_no_bias(inputs, kernel_size=kernel_size, strides=strides, padding="same", name=name + "REPARAM_1_")
    else:
        dw_1 = conv2d_no_bias(inputs, out_channel, kernel_size=kernel_size, strides=strides, padding="same", groups=groups, name=name + "REPARAM_1_")
    dw_1 = batchnorm_with_activation(dw_1, epsilon=BATCH_NORM_EPSILON, name=name + "REPARAM_1_")

    if kernel_size > 1:
        if use_depthwise:
            dw_2 = depthwise_conv2d_no_bias(inputs, 1, strides=strides, name=name + "REPARAM_2_")
        else:
            dw_2 = conv2d_no_bias(inputs, out_channel, 1, strides=strides, groups=groups, name=name + "REPARAM_2_")
        dw_2 = batchnorm_with_activation(dw_2, epsilon=BATCH_NORM_EPSILON, name=name + "REPARAM_2_")
        out = layers.Add(name=name + "REPARAM_out")([dw_1, dw_2, short] if use_shortcut else [dw_1, dw_2])
    else:
        out = layers.Add(name=name + "REPARAM_out")([dw_1, short]) if use_shortcut else dw_1
    return out


def rep_downsample_block(inputs, out_channel, strides=2, deploy=False, name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    if deploy:
        return conv2d_no_bias(inputs, out_channel, 7, strides=strides, use_bias=True, padding="same", groups=input_channel, name=name + "REPARAM_1_")

    dw_1 = conv2d_no_bias(inputs, out_channel, kernel_size=7, strides=strides, padding="same", groups=input_channel, name=name + "REPARAM_1_")
    dw_1 = batchnorm_with_activation(dw_1, epsilon=BATCH_NORM_EPSILON, name=name + "REPARAM_1_")
    dw_2 = conv2d_no_bias(inputs, out_channel, kernel_size=3, strides=strides, padding="same", groups=input_channel, name=name + "REPARAM_2_")
    dw_2 = batchnorm_with_activation(dw_2, epsilon=BATCH_NORM_EPSILON, name=name + "REPARAM_2_")
    return layers.Add(name=name + "REPARAM_out")([dw_1, dw_2])


def rep_conditional_positional_encoding(inputs, kernel_size=7, deploy=False, name=""):
    dw_1 = depthwise_conv2d_no_bias(inputs, kernel_size=kernel_size, strides=1, use_bias=True, padding="same", name=name + "REPARAM_1_")
    return dw_1 if deploy else layers.Add(name=name + "REPARAM_out")([dw_1, inputs])


def mixer_mlp_block(inputs, out_channel, mlp_ratio=3, use_attn=False, kernel_size=3, drop_rate=0, layer_scale=-1, deploy=False, activation="gelu", name=None):
    # print(f"{inputs.shape = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    layer_scale = 0 if deploy else layer_scale  # Force skip layer_scale if deploy, as it will be fused
    norm_mixer = batchnorm_with_activation(inputs, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "mixer_")
    if use_attn:
        mixer = norm_mixer if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(norm_mixer)
        mixer = multi_head_self_attention(mixer, num_heads=input_channel // 32, qkv_bias=False, out_bias=True, name=name + "attn_")
        mixer = mixer if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(mixer)
    else:
        deep_mixer = rep_conv_block(inputs, out_channel, kernel_size=kernel_size, groups=out_channel, deploy=deploy, name=name + "mixer_")
        mixer = deep_mixer if deploy else layers.Subtract(name=name + "REPARAM_TWICE_out")([deep_mixer, norm_mixer])

    if use_attn or not deploy:
        mixer = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=channel_axis, name=name + "1_gamma")(mixer) if layer_scale > 0 else mixer
        mixer = drop_block(mixer, drop_rate=drop_rate, name=name + "1_")
        mixer = layers.Add(name=name + ("1_out" if use_attn else "REPARAM_THIRD_out"))([mixer, inputs])

    mlp = conv2d_no_bias(mixer, out_channel, kernel_size=7, strides=1, use_bias=deploy, padding="same", groups=input_channel, name=name + "mlp_pre_")
    mlp = mlp if deploy else batchnorm_with_activation(mlp, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "mlp_pre_")
    mlp = mlp_block(mlp, int(out_channel * mlp_ratio), use_conv=True, activation=activation, name=name + "mlp_")

    mlp = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=channel_axis, name=name + "2_gamma")(mlp) if layer_scale > 0 else mlp
    mlp = drop_block(mlp, drop_rate=drop_rate, name=name + "2_")
    return layers.Add(name=name + "2_output")([mlp, mixer])


def FastViT(
    num_blocks=[2, 2, 4, 2],
    out_channels=[48, 96, 192, 384],
    block_types=["conv", "conv", "conv", "conv"],
    mlp_ratio=3,
    stem_width=-1,
    layer_scale=1e-5,
    input_shape=(256, 256, 3),
    deploy=False,  # build model with rep_xxx / conv+bn all being fused
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="fastvit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = rep_conv_block(inputs, stem_width, kernel_size=3, strides=2, deploy=deploy, name="stem_1_")
    nn = activation_by_name(nn, activation=activation, name="stem_1_")
    nn = rep_conv_block(nn, stem_width, kernel_size=3, strides=2, groups=stem_width, deploy=deploy, name="stem_2_")
    nn = activation_by_name(nn, activation=activation, name="stem_2_")
    nn = rep_conv_block(nn, stem_width, kernel_size=1, strides=1, deploy=deploy, name="stem_3_")
    nn = activation_by_name(nn, activation=activation, name="stem_3_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, block_type) in enumerate(zip(num_blocks, out_channels, block_types)):
        stack_name = "stack{}_".format(stack_id + 1)
        use_attn = False if block_type[0].lower() == "c" else True
        if stack_id > 0:
            nn = rep_downsample_block(nn, out_channel, deploy=deploy, name=stack_name + "downsample_")  # [???] activation is ignored
            nn = rep_conv_block(nn, out_channel, kernel_size=1, deploy=deploy, name=stack_name + "downsample_2_")
            nn = activation_by_name(nn, activation=activation, name=stack_name + "downsample_")

        if use_attn:
            nn = rep_conditional_positional_encoding(nn, deploy=deploy, name=stack_name + "cpe_")

        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            global_block_id += 1
            nn = mixer_mlp_block(
                nn, out_channel, mlp_ratio, use_attn, drop_rate=block_drop_rate, layer_scale=layer_scale, deploy=deploy, activation=activation, name=name
            )

    if num_classes > 0:
        nn = rep_conv_block(nn, out_channels[-1] * 2, kernel_size=3, strides=1, groups=out_channels[-1], deploy=deploy, name="features_")
        nn = se_module(nn, se_ratio=0.0625, divisor=1, activation="relu", name="features_se_")
        nn = activation_by_name(nn, activation=activation, name="features_")

        nn = layers.GlobalAveragePooling2D()(nn)
        nn = layers.Dropout(dropout, name="head_drop")(nn) if dropout > 0 else nn
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(nn)
    model = models.Model(inputs, nn, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "fastvit", pretrained)

    add_pre_post_process(model, rescale_mode="torch")
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


def switch_to_deploy(model):
    from keras_cv_attention_models.model_surgery.model_surgery import fuse_reparam_blocks, convert_to_fused_conv_bn_model, fuse_channel_affine_to_conv_dense

    new_model = convert_to_fused_conv_bn_model(model)
    new_model = fuse_reparam_blocks(new_model, output_layer_key="REPARAM_out")
    new_model = fuse_reparam_blocks(new_model, output_layer_key="REPARAM_TWICE_out")
    new_model = fuse_channel_affine_to_conv_dense(new_model)
    new_model = fuse_reparam_blocks(new_model, output_layer_key="REPARAM_THIRD_out")
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=model.decode_predictions)
    return new_model


@register_model
def FastViT_T8(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    return FastViT(**locals(), model_name="fastvit_t8", **kwargs)


@register_model
def FastViT_T12(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    num_blocks = [2, 2, 6, 2]
    out_channels = [64, 128, 256, 512]
    return FastViT(**locals(), model_name="fastvit_t12", **kwargs)


@register_model
def FastViT_S12(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    num_blocks = [2, 2, 6, 2]
    out_channels = [64, 128, 256, 512]
    mlp_ratio = 4
    return FastViT(**locals(), model_name="fastvit_s12", **kwargs)


@register_model
def FastViT_SA12(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    num_blocks = [2, 2, 6, 2]
    out_channels = [64, 128, 256, 512]
    mlp_ratio = 4
    block_types = ["conv", "conv", "conv", "transformer"]
    return FastViT(**locals(), model_name="fastvit_sa12", **kwargs)


@register_model
def FastViT_SA24(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    num_blocks = [4, 4, 12, 4]
    out_channels = [64, 128, 256, 512]
    mlp_ratio = 4
    block_types = ["conv", "conv", "conv", "transformer"]
    return FastViT(**locals(), model_name="fastvit_sa24", **kwargs)


@register_model
def FastViT_SA36(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    num_blocks = [6, 6, 18, 6]
    out_channels = [64, 128, 256, 512]
    mlp_ratio = 4
    layer_scale = kwargs.pop("layer_scale", 1e-6)
    block_types = ["conv", "conv", "conv", "transformer"]
    return FastViT(**locals(), model_name="fastvit_sa36", **kwargs)


@register_model
def FastViT_MA36(input_shape=(256, 256, 3), num_classes=1000, deploy=False, activation="gelu", classifier_activation="softmax", pretrained="distill", **kwargs):
    num_blocks = [6, 6, 18, 6]
    out_channels = [76, 152, 304, 608]
    mlp_ratio = 4
    layer_scale = kwargs.pop("layer_scale", 1e-6)
    block_types = ["conv", "conv", "conv", "transformer"]
    return FastViT(**locals(), model_name="fastvit_ma36", **kwargs)



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model
from keras_cv_attention_models.models import register_model


def EVA02(mlp_ratio=4 * 2 / 3, layer_scale=0, use_abs_pos_emb=True, use_rot_pos_emb=True, use_gated_mlp=True, activation="swish", model_name="eva02", **kwargs):
    kwargs.pop("kwargs", None)
    patch_size = kwargs.pop("patch_size", 14)
    force_reload_mismatch = patch_size != 14  # If patch_size not 14, force reload pos_emb and stem_conv weights
    return Beit(**locals(), **kwargs)


@register_model
def EVA02TinyPatch14(input_shape=(336, 336, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="mim_in22k_ft1k", **kwargs):
    embed_dim = 192
    depth = 12
    num_heads = 3
    return EVA02(**locals(), model_name="eva02_tiny_patch14", **kwargs)


@register_model
def EVA02SmallPatch14(input_shape=(336, 336, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="mim_in22k_ft1k", **kwargs):
    embed_dim = 384
    depth = 12
    num_heads = 6
    return EVA02(**locals(), model_name="eva02_small_patch14", **kwargs)


@register_model
def EVA02BasePatch14(
    input_shape=(448, 448, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="mim_in22k_ft22k_ft1k", **kwargs
):
    embed_dim = 768
    depth = 12
    num_heads = 12
    use_norm_mlp = True  # scale_mlp = True
    return EVA02(**locals(), model_name="eva02_base_patch14", **kwargs)


@register_model
def EVA02LargePatch14(
    input_shape=(448, 448, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="mim_m38m_ft22k_ft1k", **kwargs
):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    use_norm_mlp = True  # scale_mlp = True
    return EVA02(**locals(), model_name="eva02_large_patch14", **kwargs)



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model
from keras_cv_attention_models.models import register_model


def MetaTransformer(
    use_patch_bias=False,
    use_pre_norm=True,
    use_abs_pos_emb=True,
    attn_qv_bias=False,
    attn_qkv_bias=True,
    use_mean_pooling_head=False,
    layer_scale=0,
    model_name="meta_transformer",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    return Beit(**locals(), **kwargs)


@register_model
def MetaTransformerBasePatch16(
    input_shape=(384, 384, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="laion_2b", **kwargs
):
    depth = 12
    embed_dim = 768
    num_heads = 12
    patch_size = kwargs.pop("patch_size", 16)
    force_reload_mismatch = patch_size != 16  # If patch_size not match, force reload pos_emb and stem_conv weights
    return MetaTransformer(**locals(), model_name="meta_transformer_base_patch16", **kwargs)


@register_model
def MetaTransformerLargePatch14(
    input_shape=(336, 336, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="laion_2b", **kwargs
):
    depth = 24
    embed_dim = 1024
    num_heads = 16
    patch_size = kwargs.pop("patch_size", 14)
    force_reload_mismatch = patch_size != 14  # If patch_size not match, force reload pos_emb and stem_conv weights
    return MetaTransformer(**locals(), model_name="meta_transformer_large_patch14", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    layer_norm,
    mlp_block,
    # HeadInitializer,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "inceptionnext_base": {"imagenet": {224: "9b2bc220f5fc1cc9715daa9c899bff0b", 384: "c1740ea8e9115612eb12ce54c035ed32"}},
    "inceptionnext_small": {"imagenet": "046ec2140cce1620fd94cd1b208a9bb6"},
    "inceptionnext_tiny": {"imagenet": "1026a14bc0367f9d29a8d5bf952ac056"},
}


def inception_dwconv2d(inputs, square_kernel_size=3, band_kernel_size=11, branch_ratio=0.125, name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]

    split_size = int(input_channel * branch_ratio)
    nn_short, nn_hw, nn_ww, nn_hh = functional.split(inputs, [-1, split_size, split_size, split_size], axis=channel_axis)

    nn_hw = depthwise_conv2d_no_bias(nn_hw, kernel_size=square_kernel_size, padding="same", use_bias=True, name=name + "hw_")
    nn_ww = depthwise_conv2d_no_bias(nn_ww, kernel_size=(1, band_kernel_size), padding="same", use_bias=True, name=name + "ww_")
    nn_hh = depthwise_conv2d_no_bias(nn_hh, kernel_size=(band_kernel_size, 1), padding="same", use_bias=True, name=name + "hh_")

    out = functional.concat([nn_short, nn_hw, nn_ww, nn_hh], axis=channel_axis)
    return out


def mixer_mlp_block(inputs, mlp_ratio=4, layer_scale=0.1, drop_rate=0, activation="gelu", name=""):
    mixer = inception_dwconv2d(inputs, name=name + "mixer_")
    mixer = batchnorm_with_activation(mixer, activation=None, name=name + "mixer_")

    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]  # Channels_last only
    mlp = mlp_block(mixer, input_channel * mlp_ratio, use_conv=True, use_bias=True, activation=activation, name=name)
    return add_with_layer_scale_and_drop_block(inputs, mlp, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "output_")


def InceptionNeXt(
    num_blocks=[3, 3, 9, 3],
    embed_dims=[96, 192, 384, 768],
    mlp_ratios=[4, 4, 4, 3],
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    dropout=0,
    layer_scale=1e-6,
    classifier_activation="softmax",
    pretrained=None,
    model_name="inceptionnext",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = conv2d_no_bias(inputs, embed_dims[0], kernel_size=4, strides=4, padding="valid", use_bias=True, name="stem_")
    nn = batchnorm_with_activation(nn, activation=None, name="stem_")

    """ stacks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, embed_dim) in enumerate(zip(num_blocks, embed_dims)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = batchnorm_with_activation(nn, activation=None, name=stack_name + "downsample_")  # Using epsilon=1e-5
            nn = conv2d_no_bias(nn, embed_dim, 2, strides=2, padding="valid", use_bias=True, name=stack_name + "downsample_")

        mlp_ratio = mlp_ratios[stack_id] if isinstance(mlp_ratios, (list, tuple)) else mlp_ratios
        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = mixer_mlp_block(nn, mlp_ratio, layer_scale, block_drop_rate, activation=activation, name=name)
            global_block_id += 1

    """  Output head """
    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        head_init = initializers.TruncatedNormal(stddev=0.02)  # HeadInitializer(scale=0.02)
        nn = layers.Dense(embed_dims[-1] * 3, use_bias=True, kernel_initializer=head_init, bias_initializer="zeros", name="pre_head_dense")(nn)
        nn = activation_by_name(nn, activation=activation, name="pre_head_")
        nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="pre_head_")

        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(
            num_classes, dtype="float32", activation=classifier_activation, kernel_initializer=head_init, bias_initializer="zeros", name="predictions"
        )(nn)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "inceptionnext", pretrained)
    return model


@register_model
def InceptionNeXtTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return InceptionNeXt(**locals(), model_name="inceptionnext_tiny", **kwargs)


@register_model
def InceptionNeXtSmall(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    return InceptionNeXt(**locals(), model_name="inceptionnext_small", **kwargs)


@register_model
def InceptionNeXtBase(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    embed_dims = [128, 256, 512, 1024]
    return InceptionNeXt(**locals(), model_name="inceptionnext_base", **kwargs)



from keras_cv_attention_models.model_surgery.model_surgery import (
    SAMModel,
    DropConnect,
    add_l2_regularizer_2_model,
    align_pyramide_feature_output_by_image_data_format,
    change_model_input_shape,
    convert_to_dynamic_input_shape,
    convert_dense_to_conv,
    convert_extract_patches_to_conv,
    convert_gelu_to_approximate,
    convert_gelu_and_extract_patches_for_tflite,  # [Deprecated], use convert_gelu_to_approximate -> convert_extract_patches_to_conv instead
    convert_groups_conv2d_2_split_conv2d,
    convert_to_mixed_float16,
    convert_mixed_float16_to_float32,
    convert_to_fixed_batch_size,
    convert_to_fused_conv_bn_model,
    convert_to_token_label_model,
    convert_layers_to_deploy_inplace,
    count_params,
    export_onnx,
    fuse_sequential_conv_strict,
    fuse_channel_affine_to_conv_dense,
    fuse_reparam_blocks,
    fuse_distill_head,
    get_actual_survival_probabilities,
    get_actual_drop_connect_rates,
    get_flops,
    get_global_avg_pool_layer_id,
    get_pyramide_feature_layers,
    prepare_for_tflite,
    remove_layer_single_input,
    replace_ReLU,
    replace_add_with_drop_connect,
    replace_add_with_stochastic_depth,
    replace_stochastic_depth_with_add,
    split_model_to_head_body_tail_by_blocks,
    swin_convert_pos_emb_mlp_to_MlpPairwisePositionalEmbedding_weights,
)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    ChannelAffine,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    MixupToken,
    mlp_block,
    scaled_dot_product_attention,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

BATCH_NORM_DECAY = 0.9
BATCH_NORM_EPSILON = 1e-5
LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "uniformer_base_32": {"token_label": {224: "992e2534c5741b2fb4f35a99b40b9c98"}},
    "uniformer_base_64": {"imagenet": {224: "e852a51824f01ef9a87792395d0d8820"}, "token_label": {224: "f72f7683bfe19854c79f9395f31bdb23"}},
    "uniformer_small_32": {"token_label": {224: "efb9e6531fcc2b560abdddc16b7e8297"}},
    "uniformer_small_64": {"imagenet": {224: "867648a1d96af15ef553337e27b53ede"}, "token_label": {224: "4d8d4f874b5bcf47594587800ff61fbd"}},
    "uniformer_small_plus_32": {"imagenet": {224: "7796cce29b5ea6572330547ba7eb5e0d"}, "token_label": {224: "b1d32f5e5714b66d76ef2fecce636dfb"}},
    "uniformer_small_plus_64": {"imagenet": {224: "7d10381f4527496adb2d39c4a665c808"}, "token_label": {224: "15d6af207a0f09957a5534ae1ad540ed"}},
    "uniformer_large_64": {"token_label": {224: "b1020b4e8029209a326e8fe7183d7d28", 384: "809ba104d43e905d5b24a8ec6ee02bdd"}},
}


def multi_head_self_attention(
    inputs, num_heads=4, key_dim=0, out_shape=None, pos_emb=None, out_weight=True, qkv_bias=False, out_bias=False, attn_dropout=0, output_dropout=0, name=None
):
    input_channels = inputs.shape[-1]
    blocks = inputs.shape[1:-1]
    key_dim = key_dim if key_dim > 0 else input_channels // num_heads
    out_shape = input_channels if out_shape is None or not out_weight else out_shape
    qk_out = num_heads * key_dim
    vv_dim = out_shape // num_heads

    qkv = functional.reshape(inputs, [-1, int(np.prod(blocks)), inputs.shape[-1]]) if len(inputs.shape) > 3 else inputs
    qkv = layers.Dense(qk_out * 2 + out_shape, use_bias=qkv_bias, name=name and name + "qkv")(qkv)
    query, key, value = functional.split(qkv, [qk_out, qk_out, out_shape], axis=-1)
    query = functional.transpose(functional.reshape(query, [-1, query.shape[1], num_heads, key_dim]), [0, 2, 1, 3])  #  [batch, num_heads, hh * ww, key_dim]
    key = functional.transpose(functional.reshape(key, [-1, key.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  # [batch, num_heads, key_dim, hh * ww]
    value = functional.transpose(functional.reshape(value, [-1, value.shape[1], num_heads, vv_dim]), [0, 2, 1, 3])  # [batch, num_heads, hh * ww, vv_dim]
    # qkv = functional.reshape(qkv, [-1, np.prod(qkv.shape[1:-1]), 3, num_heads, key_dim])
    # query, key, value = functional.transpose(qkv, [2, 0, 3, 1, 4])
    # key = functional.transpose(key, [0, 1, 3, 2])

    output_shape = [*blocks, out_shape]
    attention_output = scaled_dot_product_attention(query, key, value, output_shape, pos_emb, out_weight, out_bias=out_bias, dropout=attn_dropout, name=name)
    attention_output = layers.Dropout(output_dropout, name=name and name + "out_drop")(attention_output) if output_dropout > 0 else attention_output
    return attention_output


def attn_block(
    inputs, out_channel, num_heads=0, qkv_bias=True, mlp_ratio=4, mlp_drop_rate=0, attn_drop_rate=0, drop_rate=0, gamma=-1, activation="gelu", name=""
):
    pos_emb = depthwise_conv2d_no_bias(inputs, kernel_size=3, padding="same", use_bias=True, name=name + "pos_emb_")
    pos_out = layers.Add()([inputs, pos_emb])

    # print(f">>>> {is_conv = }, {num_heads = }")
    pre_attn = pos_out if backend.image_data_format() == "channels_last" else layers.Permute((2, 3, 1), name=name + "permute_pre")(pos_out)
    height, width = pre_attn.shape[1:-1]
    pre_attn = functional.reshape(pre_attn, [-1, height * width, pre_attn.shape[-1]])  # Using 3D for attention inputs

    attn = layers.LayerNormalization(epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "attn_ln")(pre_attn)
    attn = multi_head_self_attention(attn, num_heads, qkv_bias=qkv_bias, out_bias=True, attn_dropout=attn_drop_rate, name=name + "attn_mhsa_")
    attn_out = add_with_layer_scale_and_drop_block(pre_attn, attn, layer_scale=gamma, drop_rate=drop_rate, axis=-1, name=name + "1_")

    mlp = layers.LayerNormalization(epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "mlp_ln")(attn_out)
    mlp = mlp_block(mlp, int(out_channel * mlp_ratio), drop_rate=mlp_drop_rate, use_conv=False, activation=activation, name=name + "mlp_")
    out = add_with_layer_scale_and_drop_block(attn_out, mlp, layer_scale=gamma, drop_rate=drop_rate, axis=-1, name=name + "2_")
    out = functional.reshape(out, [-1, height, width, out.shape[-1]])  # Revert 3D to 4D
    return out if backend.image_data_format() == "channels_last" else layers.Permute((3, 1, 2), name=name + "permute_post_output")(out)


def conv_block(inputs, out_channel, mlp_ratio=4, mlp_drop_rate=0, drop_rate=0, gamma=-1, activation="gelu", name=""):
    pos_emb = depthwise_conv2d_no_bias(inputs, kernel_size=3, padding="same", use_bias=True, name=name + "pos_emb_")
    pos_out = layers.Add()([inputs, pos_emb])

    attn = layers.BatchNormalization(momentum=BATCH_NORM_DECAY, epsilon=BATCH_NORM_EPSILON, name=name + "attn_bn")(pos_out)
    attn = conv2d_no_bias(attn, out_channel, 1, use_bias=True, name=name + "attn_1_")
    attn = depthwise_conv2d_no_bias(attn, kernel_size=5, padding="same", use_bias=True, name=name + "attn_")
    attn = conv2d_no_bias(attn, out_channel, 1, use_bias=True, name=name + "attn_2_")
    attn_out = add_with_layer_scale_and_drop_block(pos_out, attn, layer_scale=gamma, drop_rate=drop_rate, name=name + "1_")

    mlp = layers.BatchNormalization(momentum=BATCH_NORM_DECAY, epsilon=BATCH_NORM_EPSILON, name=name + "mlp_bn")(attn_out)
    mlp = mlp_block(mlp, int(out_channel * mlp_ratio), drop_rate=mlp_drop_rate, use_conv=True, activation=activation, name=name + "mlp_")
    return add_with_layer_scale_and_drop_block(attn_out, mlp, layer_scale=gamma, drop_rate=drop_rate, name=name + "2_")


def stem(inputs, stem_width, use_conv_stem=False, drop_rate=0, activation="gelu", name="stem_"):
    if use_conv_stem:
        nn = conv2d_no_bias(inputs, stem_width // 2, kernel_size=3, strides=2, padding="same", use_bias=True, name=name + "1_")
        nn = layers.BatchNormalization(momentum=BATCH_NORM_DECAY, epsilon=BATCH_NORM_EPSILON, name=name + "1_bn")(nn)
        nn = activation_by_name(nn, activation, name=name)
        nn = conv2d_no_bias(nn, stem_width, kernel_size=3, strides=2, padding="same", use_bias=True, name=name + "2_")
        nn = layers.BatchNormalization(momentum=BATCH_NORM_DECAY, epsilon=BATCH_NORM_EPSILON, name=name + "2_bn")(nn)
    else:
        nn = conv2d_no_bias(inputs, stem_width, 4, strides=4, padding="valid", use_bias=True, name=name)
        nn = layers.LayerNormalization(epsilon=LAYER_NORM_EPSILON, name=name + "ln")(nn)
    nn = layers.Dropout(drop_rate) if drop_rate > 0 else nn
    return nn


def Uniformer(
    num_blocks=[3, 4, 8, 3],
    out_channels=[64, 128, 320, 512],
    head_dimension=64,
    use_conv_stem=False,
    block_types=["conv", "conv", "transform", "transform"],
    stem_width=-1,
    qkv_bias=True,
    mlp_ratio=4,
    layer_scale=-1,
    mix_token=False,
    token_label_top=False,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    mlp_drop_rate=0,
    attn_drop_rate=0,
    drop_connect_rate=0,
    dropout=0,
    classifier_activation="softmax",
    pretrained=None,
    model_name="uniformer",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ stem """
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = stem(inputs, stem_width, use_conv_stem, drop_rate=mlp_drop_rate, activation=activation, name="stem_")  # It's using mlp_drop_rate for stem

    if mix_token and token_label_top:
        scale = 8  # downsample 3 times
        mixup_token = MixupToken(scale=scale)
        bbox = mixup_token(nn)
        nn = mixup_token.do_mixup_token(nn, bbox * scale)

    """ stage [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, block_type) in enumerate(zip(num_blocks, out_channels, block_types)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            if use_conv_stem:
                nn = conv2d_no_bias(nn, out_channel, kernel_size=3, strides=2, padding="same", use_bias=True, name=stack_name + "downsample_")
                nn = layers.BatchNormalization(momentum=BATCH_NORM_DECAY, epsilon=BATCH_NORM_EPSILON, name=stack_name + "downsample_bn")(nn)
            else:
                nn = conv2d_no_bias(nn, out_channel, kernel_size=2, strides=2, use_bias=True, name=stack_name + "downsample_")
                nn = layers.LayerNormalization(epsilon=LAYER_NORM_EPSILON, name=stack_name + "downsample_ln")(nn)

        is_conv_block = True if block_type[0].lower() == "c" else False
        num_heads = 0 if is_conv_block else out_channel // head_dimension
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            if is_conv_block:
                nn = conv_block(nn, out_channel, mlp_ratio, mlp_drop_rate, block_drop_rate, layer_scale, activation, block_name)
            else:
                nn = attn_block(
                    nn, out_channel, num_heads, qkv_bias, mlp_ratio, mlp_drop_rate, attn_drop_rate, block_drop_rate, layer_scale, activation, block_name
                )
            global_block_id += 1
    nn = layers.BatchNormalization(momentum=BATCH_NORM_DECAY, epsilon=BATCH_NORM_EPSILON, name="post_bn")(nn)

    """ output """
    if token_label_top and num_classes > 0:
        # Training with label token
        nn_cls = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=None)  # Don't use softmax here
        nn_aux = layers.Dense(num_classes, name="aux_head")(nn)

        if mix_token:
            nn_aux = mixup_token.do_mixup_token(nn_aux, bbox)
            nn_aux = layers.Reshape((-1, nn_aux.shape[-1]), dtype="float32", name="aux")(nn_aux)

            left, top, right, bottom = bbox
            lam = 1 - ((right - left) * (bottom - top) / (nn_aux.shape[1] * nn_aux.shape[2]))
            lam_repeat = functional.expand_dims(functional.repeat(lam, functional.shape(nn_cls)[0], axis=0), 1)
            nn_cls = layers.Concatenate(axis=-1, dtype="float32", name="class")([nn_cls, lam_repeat])
        else:
            nn_aux = layers.Reshape((-1, nn_aux.shape[-1]), dtype="float32", name="aux")(nn_aux)
        out = [nn_cls, nn_aux]
    else:
        out = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)

    model = models.Model(inputs, out, name=model_name)
    post_process = token_label_imagenet_decode_predictions if token_label_top else None
    add_pre_post_process(model, rescale_mode="torch", post_process=post_process)
    pretrained = "token_label" if pretrained is not None and "token" in pretrained.lower() else pretrained
    reload_model_weights(model, PRETRAINED_DICT, "uniformer", pretrained)
    return model


def token_label_imagenet_decode_predictions(preds, top=5, classifier_activation="softmax", do_decode=True):
    from keras_cv_attention_models.imagenet.eval_func import decode_predictions

    preds = preds[0] + 0.5 * functional.reduce_max(preds[1], axis=1)
    preds = activation_by_name(preds, classifier_activation) if classifier_activation is not None else preds
    return decode_predictions(preds.numpy(), top=top) if do_decode else preds


@register_model
def UniformerSmall32(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="token_label", **kwargs):
    num_blocks = [3, 4, 8, 3]
    head_dimension = 32
    return Uniformer(**locals(), model_name="uniformer_small_32", **kwargs)


@register_model
def UniformerSmall64(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 8, 3]
    head_dimension = 64
    return Uniformer(**locals(), model_name="uniformer_small_64", **kwargs)


@register_model
def UniformerSmallPlus32(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="imagenet", **kwargs):
    num_blocks = [3, 5, 9, 3]
    head_dimension = 32
    use_conv_stem = True
    return Uniformer(**locals(), model_name="uniformer_small_plus_32", **kwargs)


@register_model
def UniformerSmallPlus64(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="imagenet", **kwargs):
    num_blocks = [3, 5, 9, 3]
    head_dimension = 64
    use_conv_stem = True
    return Uniformer(**locals(), model_name="uniformer_small_plus_64", **kwargs)


@register_model
def UniformerBase32(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="token_label", **kwargs):
    num_blocks = [5, 8, 20, 7]
    head_dimension = 32
    return Uniformer(**locals(), model_name="uniformer_base_32", **kwargs)


@register_model
def UniformerBase64(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="imagenet", **kwargs):
    num_blocks = [5, 8, 20, 7]
    head_dimension = 64
    return Uniformer(**locals(), model_name="uniformer_base_64", **kwargs)


@register_model
def UniformerLarge64(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", token_label_top=False, pretrained="token_label", **kwargs):
    num_blocks = [5, 10, 24, 7]
    out_channels = [128, 192, 448, 640]
    head_dimension = 64
    layer_scale = 1e-6
    return Uniformer(**locals(), model_name="uniformer_large_64", **kwargs)
